import math

import numpy as np


class BlackOil:
    """Black Oil Stream

    Set a condition and calculate density, viscosity or compressibility
    """

    def __init__(self, oil_api: float, bubblepoint: float, gas_sg: float) -> None:
        """Initialize a Black Oil Stream

        Args:
            oil_api (float): Oil API, 10 to 40
            bubblepoint (float): Bubble Point Pressure, PSIG
            gas_sg (float): In-Situ Gas Specific Gravity

        Returns:
            Self
        """
        api_min = 10
        api_max = 40
        if (api_min < oil_api < api_max) is False:
            raise ValueError(f"Oil API {oil_api} Outside Range of {api_min} to {api_max}")

        bub_min = 1000
        bub_max = 3000
        if (bub_min < bubblepoint < bub_max) is False:
            raise ValueError(f"Bubblepoint {bubblepoint} Outside Range of {bub_min} to {bub_max}")

        sg_min = 0.5
        sg_max = 1.2
        if (sg_min < gas_sg < sg_max) is False:
            raise ValueError(f"Gas SG {gas_sg} Outside Range of {sg_min} to {sg_max}")

        self.oil_api = oil_api
        self.pbp = bubblepoint
        self.gas_sg = gas_sg

    def __repr__(self):
        return f"Oil: {self.oil_api} API and a {round(self.gas_sg, 2)} SG Gas"

    @classmethod
    def schrader(cls):
        """Schrader Bluff Black Oil

        Generic Schrader Bluff black oil with preset properties

        Args:
            oil_api (float): 22 API
            bubblepoint (float): 1750 psig
            gas_sg (float): 0.65
        """
        return cls(oil_api=22, bubblepoint=1750, gas_sg=0.65)

    @classmethod
    def kuparuk(cls):
        """Kuparuk Black Oil

        Generic Kuparuk black oil with preset properties

        Args:
            oil_api (float): 24 API
            bubblepoint (float): 2250 psig
            gas_sg (float): 0.65
        """
        return cls(oil_api=24, bubblepoint=2250, gas_sg=0.65)

    @classmethod
    def sag(cls):
        """Sagavanirktok Black Oil

        Generic Sagavanirktok black oil with preset properties

        Args:
            oil_api (float): 28 API
            bubblepoint (float): 2700 psig
            gas_sg (float): 0.65
        """
        return cls(oil_api=28, bubblepoint=2700, gas_sg=0.65)

    @classmethod
    def ugnu(cls):
        """Ugnu Black Oil

        Generic Ugnu black oil with preset properties

        Args:
            oil_api (float): 14 API
            bubblepoint (float): 1500 psig
            gas_sg (float): 0.65
        """
        return cls(oil_api=14, bubblepoint=1500, gas_sg=0.65)

    @classmethod
    def test_oil(cls):
        """Test Oil

        Used to run tests on the black oil methods. Uses 0.55 gas
        specific gravity with is the same as pure methane.
        """
        return cls(oil_api=22, bubblepoint=1750, gas_sg=0.55)

    def condition(self, press: float, temp: float):
        """Set condition of evaluation

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F

        Returns:
            self
        """
        self.press = press
        self.temp = temp
        return self

    def gas_solubility(self) -> float:
        """Gas Solubility in Oil (Solution GOR)

        Return the gas solubility in the oil, scf/stb.

        Args:
            None

        Returns:
            rs (float): Gas solubility in the oil, scf/stb
        """
        if self.press > self.pbp:  # above bubblepoint pressure
            press = self.pbp  # calculate gas solubility using bubblepoint pressure
        else:
            press = self.press
        rs = self.solubility_kartoatmodjo(press, self.temp, self.oil_api, self.gas_sg)
        return rs

    @property
    def compress(self) -> float:
        """Oil Compressibility Isothermal

        Calculate isothermal oil compressibility.

        Args:
            None

        Returns:
            co (float): oil compressibility, psi**-1
        """
        rs = self.gas_solubility()  # solubility of gas in the oil
        if self.press > self.pbp:  # above bubblepoint
            co = self.compressibility_vasquez_above(self.press, self.temp, self.oil_api, self.gas_sg, rs)
        else:  # below the bubblepoint
            co = self.compressibility_mccain_below(self.press, self.temp, self.oil_api, self.gas_sg, rs)
            # co = self.compressibility_kartoatmodjo_above(self.press, self.temp, self.oil_api, self.gas_sg, rs)
        return co

    def oil_fvf(self) -> float:
        """Oil Formation Volume Factor, Bo

        Calculate formation volume factor for the oil phase.

        Args:
            None

        Returns:
            bo (float): oil formation volume, rb/stb
        """
        rs = self.gas_solubility()
        bo = self.fvf_kartoatmodjo_below(self.temp, self.oil_api, self.gas_sg, rs)
        if self.press > self.pbp:  # above bubblepoint pressure
            bob = bo
            co = self.compress
            bo = self.fvf_vasquez_above(self.press, self.pbp, bob, co)
        return bo

    @property  # property decorator makes it so you don't need brackets
    def density(self) -> float:
        """Live oil density, lbm/ft3

        Calculate the live density of the oil.

        Args:
            None

        Returns:
            rho_oil (float): density of the oil, lbm/ft3
        """
        rs = self.gas_solubility()
        bo = self.oil_fvf()
        rho_oil = self.live_oil_density(self.oil_api, self.gas_sg, rs, bo)
        return rho_oil

    @property
    def viscosity(self) -> float:
        """Live oil viscosity, cP

        Return the live oil viscosity, cP.
        Requires pressure and temperature condition to previously be set.

        Args:
            None

        Returns:
            uoil (float): live oil viscosity, cP
        """
        uod = self.viscosity_dead_kartoatmodjo(self.temp, self.oil_api)
        uol = self.viscosity_live_kartoatmodjo_below(uod, self.gas_solubility())
        if self.press > self.pbp:  # above bubblepoint
            uob = uol
            uol = self.viscosity_live_kartoatmodjo_above(uob, self.press, self.pbp)
        return uol

    @property
    def tension(self) -> float:
        """Live Oil Surface Tension

        Calculate the live oil surface tension

        Args:
            None

        Returns:
            sigo (float): Live Oil Surface Tension, lbf/ft"""
        sigod = self.tension_dead_abdul(self.temp, self.oil_api)
        sigol = self.tension_live_abdul(sigod, self.gas_solubility())  # dyne/cm
        sigo = sigol * 0.0000685  # lbf/ft
        return sigo

    @staticmethod
    def fvf_standing_below(temp: float, oil_api: float, gas_sg: float, rs: float) -> float:
        """Standing Oil Formation Volume Factor

        Calculate formation volume factor for the oil phase.
        Uses the Standing California 1980 correlation.

        Args:
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, air
            rs (float): Solubility of Gas in Oil, SCF/STB

        Returns:
            bo (float): Oil FVF at / below bubblepoint, rb/stb
        """
        oil_sg = 141.5 / (oil_api + 131.5)  # oil specific gravity
        bo = 0.972 + 1.47 * 10**-4 * (rs * (gas_sg / oil_sg) ** 0.5 + 1.25 * temp) ** 1.175
        return bo

    @staticmethod
    def fvf_almarhoun_below(temp: float, oil_api: float, gas_sg: float, rs: float) -> float:
        """Al-Marhoun FVF Below Bubblepoint

        Calculate the formation volume factor at / or below bubblepoint.
        The is method is different than the 1985 method commonly referenced.

        Args:
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in Oil, SCF/STB

        Returns:
            bo (float): Oil FVF at / below bubblepoint, rb/stb

        References:
            - New Correlations for FVF of Oil and Gas, M. Al-Marhoun (1992) PETSOC-92-03-02
            - Black Oil Property Correlations State of the Art, M. Al-Marhoun (2015) SPE-172833-MS
        """
        a1 = 0.177342 * 10**-3
        a2 = 0.220163 * 10**-3
        a3 = 4.292580 * 10**-6
        a4 = 0.528707 * 10**-3
        oil_sg = 141.5 / (oil_api + 131.5)  # oil specific gravity
        bo = 1 + a1 * rs + a2 * rs * gas_sg / oil_sg + a3 * rs * (temp - 60) * (1 - oil_sg) + a4 * (temp - 60)
        return bo

    @staticmethod
    def fvf_kartoatmodjo_below(temp: float, oil_api: float, gas_sg: float, rs: float) -> float:
        """Kartoatmodjo and Schmidt FVF Below Bubblepoint

        Calculate the formation volume factor at / or below bubblepoint.
        Correlation uses gas_sg at 100 psig, which is the composition of the free gas
        released at 100 psig. Assumption is made that the difference in composition of the
        free gas between 100 psig and 0 psig is negligible.

        Args:
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in Oil, SCF/STB

        Returns:
            bo (float): Oil FVF at / below Bubblepoint, rb/stb

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 285
        """
        oil_sg = 141.5 / (oil_api + 131.5)  # oil specific gravity
        f = rs**0.755 * gas_sg**0.25 * oil_sg**-1.5 + 0.45 * temp
        bo = 0.98496 + 0.0001 * f**1.5
        return bo

    @staticmethod
    def fvf_vasquez_above(press: float, pbp: float, bob: float, co: float) -> float:
        """Vasquez and Beggs FVF Above Bubblepoint

        Calculate formation volume factor above bubblepoint:

        Args:
            press (float): Pressure of the oil, psig
            pbp (float): Bubblepoint Pressure, psig
            bob (float): Oil FVF at Bubblepoint, rb/stb
            co (float): Oil Isothermal Compressibility, psi-1

        Returns:
            bo (float): Oil FVF above bubblepoint, rb/stb

        References:
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 286
        """
        bo = bob * math.exp(co * (pbp - press))
        return bo

    @staticmethod
    def solubility_kartoatmodjo(press: float, temp: float, oil_api: float, gas_sg: float) -> float:
        """Kartoatmodjo and Schmidt Solubility of Gas in Oil

        Calculate the Solubility of Gas in Oil.
        Correlation uses gas_sg at 100 psig, which is the composition of the free gas
        released at 100 psig. Assumption is made that the difference in composition of the
        free gas between 100 psig and 0 psig is negligible.

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air

        Returns:
            rs (float): Solubility of the Gas in the Oil, scf/stb

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 285
        """
        if oil_api <= 30:
            c1, c2, c3, c4 = [0.05958, 0.7972, 1.0014, 13.1405]
        else:  # oilAPI greater than 30API
            c1, c2, c3, c4 = [0.0315, 0.7587, 1.0937, 11.289]
        pabs = press + 14.7  # absolute pressure
        # print(pabs)
        with np.errstate(invalid="raise"):
            rs = c1 * gas_sg**c2 * pabs**c3 * 10 ** (c4 * oil_api / (temp + 460))
        # print(rs)
        return rs

    @staticmethod
    def solubility_vasquez(press: float, temp: float, oil_api: float, gas_sg: float) -> float:
        """Vasquez and Beggs Solubility of Gas in Oil

        Calculate solubility of the gas in oil. Using the Vasquez and Beggs Methodology

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air

        Returns:
            rs (float): Solubility of the Gas in the Oil, scf/stb

        References:
            - Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 22
            - Correlations for Fluid Physical Property... Vasquez and Beggs (1980)
        """
        if oil_api <= 30:
            c1, c2, c3 = [0.0362, 1.0937, 25.7240]
        else:  # if oilAPI is greater than 30API
            c1, c2, c3 = [0.0178, 1.187, 23.9310]
        pabs = press + 14.7  # absolute pressure
        rs = c1 * gas_sg * pabs**c2 * math.exp(c3 * oil_api / (temp + 460))
        return rs

    @staticmethod
    def compressibility_vasquez_above(press: float, temp: float, oil_api: float, gas_sg: float, rs: float) -> float:
        """Vasquez and Beggs compressibility of oil above bubblepoint

        Calculate isothermal oil compressibility. Inverse of the bulk modulus of elasticity.
        Valid for pressures that are above the oil bubblepoint.

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in the oil, scf/stb

        Returns:
            co (float): oil compressibility, psi**-1

        References:
            - Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 22
            - Correlations for Fluid Physical Property... Vasquez and Beggs (1980)
        """
        pabs = press + 14.7
        co = (5 * rs + 17.2 * temp - 1180 * gas_sg + 12.61 * oil_api - 1433) / (pabs * 10**5)
        return co

    @staticmethod
    def compressibility_kartoatmodjo_above(
        press: float, temp: float, oil_api: float, gas_sg: float, rs: float
    ) -> float:
        """Kartoatmodjo compressibility of oil above bubblepoint

        Calculate isothermal oil compressibility. Valid for all (?) pressures.
        Error in the paper, oil_api is ommitted and temp coef is not 0.7606.
        Textbook has the correct equation, verified via internets.

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in the oil, scf/stb

        Returns:
            co (float): oil compressibility, psi**-1

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 286
        """
        pa = press + 14.7
        co = 6.8257 * 10**-6 * rs**0.5002 * oil_api**0.3613 * temp**0.7606 * gas_sg**-0.35505 / pa
        return co

    @staticmethod
    def compressibility_mccain_below(press: float, temp: float, oil_api: float, gas_sg: float, rs: float) -> float:
        """McCain et. al compressibility of oil below bubblepoint

        Calculate isothermal oil compressibility. Valid for pressures below bubblepoint.
        Equation 5 in the referenced paper by McCain.

        Args:
            press (float): Pressure of the oil, psig
            temp (float): Temperature of the oil, deg F
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in the oil, scf/stb

        Returns:
            co (float): oil compressibility, psi**-1

        References:
            - Coefficient of Isothermal Compressibility of Black Oil, McCain, Rollins, Villena (1988) SPE-15664-PA
        """
        # gas_sg isn't used, eqn 4 in paper takes oil bubblepoint as input
        pa = press + 14.7
        co = math.exp(
            -7.633 - 1.497 * math.log(pa) + 1.115 * math.log(temp) + 0.533 * math.log(oil_api) + 0.184 * math.log(rs)
        )
        return co

    @staticmethod
    def live_oil_density(oil_api: float, gas_sg: float, rs: float, bo: float) -> float:
        """Live Oil Density, lbm/ft3

        Calculate the live density of the oil.

        Args:
            oil_api (float): Oil API Degrees
            gas_sg (float): Gas Specific Gravity, relative to air
            rs (float): Solubility of Gas in the oil, scf/stb
            bo (float): Oil FVF, rb/stb

        Returns:
            rho_oil (float): density of the oil, lbm/ft3

        References:
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 288, Eqn B-34
        """
        oil_sg = 141.5 / (oil_api + 131.5)  # oil specific gravity
        rho_oil = (62.42796 * oil_sg + (0.0136 * gas_sg * rs)) / bo
        return rho_oil

    @staticmethod
    def viscosity_dead_beggs(temp: float, oil_api: float) -> float:
        """Beggs Dead Oil Viscosity

        Dead oil viscosity following Beggs methodology.

        Args:
            temp (float): Oil Temperature, deg F
            oil_api (float): Oil API Degrees

        Returns:
            uod (float): Dead Oil Viscosity, cP

        References:
            - Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 23
            - Correlations for Fluid Physical Property... Vasquez and Beggs (1980)
        """
        x = temp**-1.163 * math.exp(6.9824 - 0.04658 * oil_api)
        uod = 10**x - 1
        return uod

    @staticmethod
    def viscosity_live_beggs_below(uod: float, rs: float) -> float:
        """Live Oil Viscosity, Below Bubble Point

        Live oil viscosity below bubblepoint following Beggs methodology.

        Args:
            uod (float): Dead Oil Viscosity, cP
            rs (float): Oil API Degrees

        Returns:
            uol (float): Live Oil Viscosity, cP

        References:
            - Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 23
            - Correlations for Fluid Physical Property... Vasquez and Beggs (1980)
        """
        a = 10.715 * (rs + 100) ** -0.515
        b = 5.44 * (rs + 150) ** -0.338
        uol = a * (uod**b)
        return uol

    @staticmethod
    def viscosity_live_beggs_above(uob: float, press: float, pbp: float) -> float:
        """Live Oil Viscosity, Above Bubble Point

        Live oil viscosity above bubblepoint following Beggs methodology.

        Args:
            uob (float): Live Oil Viscosity at Bubblepoint, cP
            press (float): Pressure, psig
            pbp (float): Bubblepoint Pressure, psig

        Returns:
            uol (float): Live Oil Viscosity, cP

        References:
            - Fundamental Principles of Reservoir Engineering, B.Towler (2002) Page 23
            - Correlations for Fluid Physical Property... Vasquez and Beggs (1980)
        """
        pa = press + 14.7
        m = 2.6 * press**1.187 * math.exp(-11.513 - 8.98 * 10**-5 * press)
        uol = uob * (pa / pbp) ** m
        return uol

    @staticmethod
    def viscosity_dead_kartoatmodjo(temp: float, oil_api: float) -> float:
        """Kartoatmodjo Dead Oil Viscosity

        Dead oil viscosity following Kartoatmodjo methodology.

        Args:
            temp (float): Oil Temperature, deg F
            oil_api (float): Oil API Degrees

        Returns:
            uod (float): Dead Oil Viscosity, cP

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 289
        """
        uod = (16 * 10**8) * temp**-2.8177 * (math.log(oil_api, 10)) ** (5.7526 * math.log(temp, 10) - 26.9718)
        return uod

    @staticmethod
    def viscosity_live_kartoatmodjo_below(uod: float, rs: float) -> float:
        """Live Oil Viscosity, Below Bubble Point

        Live oil viscosity below bubblepoint following Kartoatmodjo methodology.

        Args:
            uod (float): Dead Oil Viscosity, cP
            rs (float): Gas Solubility, scf/stb

        Returns:
            uol (float): Live Oil Viscosity, cP

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 289
        """
        y = 10 ** (-0.00081 * rs)
        f = (0.2001 + 0.8428 * 10 ** (-0.000845 * rs)) * uod ** (0.43 + 0.5165 * y)
        uol = -0.06821 + 0.9824 * f + 0.0004034 * f**2
        return uol

    @staticmethod
    def viscosity_live_kartoatmodjo_above(uob: float, press: float, pbp: float) -> float:
        """Live Oil Viscosity, Above Bubble Point

        Live oil viscosity above bubblepoint following Kartoatmodjo methodology.

        Args:
            uob (float): Live Oil Viscosity at Bubblepoint, cP
            press (float): Pressure, psig
            pbp (float): Bubblepoint Pressure, psig

        Returns:
            uol (float): Live Oil Viscosity, cP

        References:
            - New Correlations for Crude Oil Physical Properties, Kartoatmodjo and Schmidt (1991) SPE-23556
            - Applied Multiphase Flow in Pipes (2017) Al-Safran and Brill, Page 290
        """
        uol = 1.00081 * uob + 0.001127 * (press - pbp) * (-0.006517 * uob**1.8148 + 0.038 * uob**1.590)
        return uol

    @staticmethod
    def tension_dead_abdul(temp: float, oil_api: float) -> float:
        """Surface Tension Dead Oil

        Dead oil surface tension from the Abdul-Majeed 2000 Method

        Args:
            temp (float): Oil Temperature, deg F
            oil_api (float): Oil API Degrees

        Returns:
            sigod (float): Dead Oil Surface Tension, dyne/cm

        References:
            - Estimation of gas - oil surface tension, G.H. Abdul-Majeed, J Petroleum Science (2000)"""

        temp_c = (5 / 9) * (temp - 32)
        A = 1.11591 - 0.00305 * temp_c
        sigod = A * (38.085 - 0.259 * oil_api)
        return sigod

    @staticmethod
    def tension_live_abdul(sigod: float, rs: float) -> float:
        """Surface Tension Live Oil

        Live oil surface tension from the Abdul-Majeed 2000 Method

        Args:
            sigod (float): Dead Oil Surface Tension, dyne/cm
            oil_api (float): Gas Solubility, scf/stb

        Returns:
            sigol (float): Live Oil Surface Tension, dyne/cm

        References:
            - Estimation of gas - oil surface tension, G.H. Abdul-Majeed, J Petroleum Science (2000)
        """
        rs_si = rs * 7.4801 / 42  # units of ft3/ft3, equivalent to m3/m3
        if rs_si < 50:
            sigol = sigod / (1 + 0.02549 * rs_si**1.0157)
        else:
            sigol = sigod * 32.0436 * rs_si**-1.1367
        return sigol



"""Jet Pump Network Solver

Add mutliple BatchPumps to a network and provide a shared resource. The shared
resource can be either lift water (power fluid) or total water.
"""

import numpy as np
import pandas as pd
from ortools.sat.python import cp_model

from woffl.assembly.batchpump import BatchPump

SCALE = 100  # CP-SAT requires integers; multiply floats by this before rounding


class WellNetwork:
    """A Well Network is a collection of multiple oil wells that share a common resource. The
    common resource is typically power fluid, but can also be total fluid"""

    def __init__(
        self,
        pwh_hdr: float | None,
        ppf_hdr: float | None,
        well_list: list[BatchPump],
        pad_name: str = "na",
    ) -> None:
        """Well Network Solver

        Used for feeding a network of multiple pumps together from a shared resource. The
        shared resource is typically power fluid, but can also be total fluid. If the well head
        header pressure or power fluid header pressure are left as None, then the well header
        pressure or power fluid pressure inside the individual BatchPumps will be used. This is
        convenient if a sensitivity is desired to look at changing power fluid pressure or well head

        Args:
            pwh_hdr (float): Pressure Wellhead Header, psig
            ppf_hdr (float): Pressure Power Fluid Header, psig
            batchlist (list): List of BatchPumps to run through
            pad_name (str): Name of the Pad or Network assessed
        """
        self.pwh_hdr = pwh_hdr
        self.ppf_hdr = ppf_hdr
        self.well_list = well_list
        self.pad_name = pad_name
        self._update_wells()
        self.results = False  # used for easily viewing if results have been processed

    def update_press(self, kind: str, psig: float) -> None:
        """Update Header Pressures

        Used to update different header pressure instead of re-initializing everything.
        Header pressures that can be updated include wellhead (production) or power fluid.

        Args:
            kind (str): Kind of Pressure to update. "wellhead" or "powerfluid".
            psig (float): Pressure to update with, psig
        """
        # chat gpt thinks using the reservoir method will fail, since you are calling ipr_su
        press_map = {"wellhead": "pwh_hdr", "powerfluid": "ppf_hdr"}

        # Validate the 'kind' argument
        if kind not in press_map:
            valid_kind = ", ".join(press_map.keys())
            raise ValueError(f"Invalid value for 'kind': {kind}. Expected {valid_kind}.")

        attr_name = press_map[kind]
        setattr(self, attr_name, psig)
        self._update_wells()

    def _update_wells(self) -> None:
        """Internal Method for Updating Well Pressures

        Can be run anytime a pressure is modified to update all of the wells in the list.
        Cascades network power fluid header pressure or well head header pressure.
        """
        if self.pwh_hdr is not None:
            for well in self.well_list:
                well.update_press("wellhead", self.pwh_hdr)

        if self.ppf_hdr is not None:
            for well in self.well_list:
                well.update_press("powerfluid", self.ppf_hdr)

    def add_well(self, well: BatchPump) -> None:
        """Add Well onto the Network"""
        self.well_list.append(well)
        self._update_wells()

    def drop_well(self, well: BatchPump) -> None:
        """Remove Well from the Network"""
        self.well_list.remove(well)

    def optimize(
        self,
        qpf_tot: float,
        water_key: str = "lift_wat",
        allow_shutin: bool = False,
    ) -> pd.DataFrame:
        """Run MCKP Optimization on Network Wells

        Args:
            qpf_tot (float): Total surface pump capacity, BWPD
            water_key (str): Column for capacity constraint, "lift_wat" or "totl_wat"
            allow_shutin (bool): If True, solver may shut in a well when its water is better used elsewhere

        Returns:
            df (DataFrame): One row per well with selected pump and rates
        """
        return optimize_jet_pumps(self.well_list, qpf_tot, water_key, allow_shutin)


def optimize_jet_pumps(
    well_list: list[BatchPump],
    qpf_tot: float,
    water_key: str = "lift_wat",
    allow_shutin: bool = False,
) -> pd.DataFrame:
    """Optimize Jet Pumps via Multiple-Choice Knapsack

    Each well picks exactly one jet pump from its semi-finalists to maximize
    total oil production subject to a shared power fluid capacity constraint.
    Uses the CP-SAT solver from ortools.

    Args:
        well_list (list[BatchPump]): Wells with batch_run() and process_results() already called
        qpf_tot (float): Total surface pump capacity, BWPD
        water_key (str): Column for capacity constraint, "lift_wat" or "totl_wat"
        allow_shutin (bool): If True, solver may shut in a well when its water is better used elsewhere

    Returns:
        df (DataFrame): One row per well with selected pump and rates

    Raises:
        ValueError: If any well has no semi-finalists
        RuntimeError: If the problem is infeasible (capacity too small)
    """
    # collect semi-finalist candidates per well
    candidates = []
    for well in well_list:
        df_semi = well.df[well.df["semi"]].reset_index(drop=True)
        if df_semi.empty:
            raise ValueError(f"Well '{well.wellname}' has no semi-finalists")
        candidates.append(df_semi)

    model = cp_model.CpModel()

    # decision variables: x[i][j] = 1 if well i selects semi-finalist j
    x = []
    for i, df_semi in enumerate(candidates):
        well_vars = [model.new_bool_var(f"w{i}_p{j}") for j in range(len(df_semi))]
        x.append(well_vars)
        if allow_shutin:
            model.add_at_most_one(well_vars)
        else:
            model.add_exactly_one(well_vars)

    # capacity constraint: total water <= qpf_tot
    capacity_scaled = int(np.floor(qpf_tot * SCALE))
    water_terms = []
    for i, df_semi in enumerate(candidates):
        for j, wat in enumerate(df_semi[water_key]):
            water_terms.append(int(np.ceil(wat * SCALE)) * x[i][j])
    model.add(sum(water_terms) <= capacity_scaled)

    # objective: maximize total oil
    oil_terms = []
    for i, df_semi in enumerate(candidates):
        for j, oil in enumerate(df_semi["qoil_std"]):
            oil_terms.append(int(np.floor(oil * SCALE)) * x[i][j])
    model.maximize(sum(oil_terms))

    # solve
    solver = cp_model.CpSolver()
    status = solver.solve(model)

    if status not in (cp_model.OPTIMAL, cp_model.FEASIBLE):
        min_water = sum(df[water_key].min() for df in candidates)
        raise RuntimeError(
            f"MCKP infeasible: {qpf_tot:.0f} bwpd capacity cannot serve all wells. "
            f"Minimum required: {min_water:.0f} bwpd."
        )

    # extract solution
    results = []
    for i, (well, df_semi) in enumerate(zip(well_list, candidates)):
        selected = False
        for j in range(len(df_semi)):
            if solver.value(x[i][j]):
                row = df_semi.iloc[j]
                results.append(
                    {
                        "wellname": well.wellname,
                        "nozzle": row["nozzle"],
                        "throat": row["throat"],
                        "qoil_std": row["qoil_std"],
                        "lift_wat": row["lift_wat"],
                        "form_wat": row["form_wat"],
                        "totl_wat": row["totl_wat"],
                    }
                )
                selected = True
                break
        if not selected:
            results.append(
                {
                    "wellname": well.wellname,
                    "nozzle": "off",
                    "throat": "off",
                    "qoil_std": 0.0,
                    "lift_wat": 0.0,
                    "form_wat": 0.0,
                    "totl_wat": 0.0,
                }
            )

    return pd.DataFrame(results)



import json
from pathlib import Path

import numpy as np
import pytest

from woffl.pvt.blackoil import BlackOil


def compute_blackoil_data(
    prs_ray: np.ndarray | list, temp: float, oil_api: float, bubblepoint: float, gas_sg: float
) -> dict:
    """Compute BlackOil Data

    Create a list of properties of a formgas. Can be used to compare to results obtained with hysys.
    Density and oil viscosity.
    """
    py_oil = BlackOil(oil_api=oil_api, bubblepoint=bubblepoint, gas_sg=gas_sg)
    rho_oil, visc_oil = [], []

    for prs in prs_ray:
        py_gas = py_oil.condition(prs, temp)

        rho_oil.append(py_gas.density)
        visc_oil.append(py_gas.viscosity)

    pyoil = {"rho_oil": rho_oil, "visc_oil": visc_oil}
    return pyoil


@pytest.fixture(scope="module")
def hysys_blackoil():
    hysys_path = Path(__file__).parents[1] / "data" / "hysys_blackoil_peng_rob.json"
    with open(hysys_path) as json_file:
        return json.load(json_file)


@pytest.fixture(scope="module")
def python_blackoil(hysys_blackoil):
    return compute_blackoil_data(
        hysys_blackoil["pres_psig"],
        hysys_blackoil["temp_degf"],
        hysys_blackoil["oil_api"],
        hysys_blackoil["bubblepoint"],
        hysys_blackoil["gas_sg"],
    )


def test_oil_density(hysys_blackoil, python_blackoil) -> None:
    np.testing.assert_allclose(hysys_blackoil["rho_oil"], python_blackoil["rho_oil"], rtol=0.05)


def test_oil_viscosity(hysys_blackoil, python_blackoil) -> None:
    # 75% error, why are we even testing...haha
    np.testing.assert_allclose(hysys_blackoil["visc_oil"], python_blackoil["visc_oil"], rtol=0.75)


def test_oil_tension() -> None:
    py_boil = BlackOil.test_oil()
    py_boil.condition(2500, 80)
    assert py_boil.tension / 0.0000685 == pytest.approx(16.04, rel=0.01)  # dyne/cm



"""Flow Folder

Collection of classes, functions and methods to describe flow inside pipes and jetpumps.
Flow equations for single phase, two phase, reservoir inflow and jet pumps are present.
"""

from .inflow import InFlow



import numpy as np
import pytest

from woffl.geometry.wellprofile import WellProfile


sch_profile = WellProfile.schrader()
# sch_profile.plot_raw()
# sch_profile.plot_filter()

kup_profile = WellProfile.kuparuk()
# kup_profile.plot_raw()
# kup_profile.plot_filter()


def test_sch_vs_kup() -> None:
    assert sch_profile.vd_ray[-1] == pytest.approx(4500, 0.1)
    assert kup_profile.vd_ray[-1] == pytest.approx(9000, 0.1)


# Sample Data
md_list = [0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000]
vd_list = [0, 95, 180, 270, 350, 375, 400, 410, 415, 420, 425]
jetpump_md = 450
wp = WellProfile(md_list, vd_list, jetpump_md)


def test_initialization():
    """Test WellProfile initialization with valid inputs"""
    assert len(wp.md_ray) == len(wp.vd_ray) == len(wp.md_ray)
    assert len(wp.md_fit) == len(wp.vd_fit) == len(wp.md_fit)
    assert wp.jetpump_md == jetpump_md


def test_initialization_invalid_lengths():
    """Test initialization failure for mismatched lengths"""
    with pytest.raises(ValueError, match="Lists for Measured Depth and Vertical Depth need to be the same length"):
        WellProfile([0, 100], [0, 80, 150], 100)


def test_initialization_invalid_order():
    """Test initialization failure for measured depth being too short"""
    with pytest.raises(ValueError, match="Measured Depth needs to extend farther than Vertical Depth"):
        WellProfile([0, 80, 150], [0, 100, 200], 100)


def test_init_invalid_jetpump() -> None:
    with pytest.raises(ValueError, match="Jet pump not inside well profile measured depth"):
        WellProfile(md_list, vd_list, jetpump_md=1200)


def test_vd_interp():
    """Test vertical depth interpolation"""
    item_md = 350
    assert wp.vd_interp(item_md) <= item_md
    assert wp.vd_interp(item_md) == pytest.approx(310, 0.05)


def test_hd_interp():
    """Test horizontal distance interpolation"""
    item_md = 350
    assert wp.hd_interp(item_md) <= item_md
    assert wp.hd_interp(item_md) == pytest.approx(160, 0.05)


def test_md_interp():
    """Test measured depth interpolation"""
    item_vd = 190
    assert wp.md_interp(item_vd) >= item_vd
    assert wp.md_interp(item_vd) == pytest.approx(210, 0.01)


def test_filter():
    """Test profile filtering reduces number of points while maintaining key structure"""
    assert len(wp.md_fit) < len(wp.md_ray)  # Filtering should reduce points
    assert wp.md_fit[0] == 0  # First point should always be 0
    assert wp.vd_fit[0] == 0  # First point should always be 0


def test_outflow_spacing():
    """Test outflow spacing generates valid segments"""
    md_seg, vd_seg = wp.outflow_spacing(50)  # 50-ft segments
    assert len(md_seg) == len(vd_seg)  # Should be same length
    assert md_seg[-1] == jetpump_md  # outflow ends at jetpump depth
    assert md_seg[-1] <= max(md_list)  # Should not exceed max depth



import json
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from woffl.pvt import BlackOil, FormGas, FormWater, ResMix


def compute_resmix_data(
    prs_ray: np.ndarray | pd.Series, temp: float, wc: float, fgor: float, oil_api: float, pbub: float, gas_sg: float
) -> dict:
    """Compute Reservoir Mixture

    Create a list of mass and volume fractions for Oil, Water and Gas
    in a mixture. Can be used to compare to results obtained with hysys
    """

    py_oil = BlackOil(oil_api=oil_api, bubblepoint=pbub, gas_sg=gas_sg)
    py_wat = FormWater(wat_sg=1)
    py_gas = FormGas(gas_sg=gas_sg)
    py_mix = ResMix(wc=wc, fgor=fgor, oil=py_oil, wat=py_wat, gas=py_gas)

    mfac_oil, mfac_wat, mfac_gas = [], [], []
    vfac_oil, vfac_wat, vfac_gas = [], [], []
    rho_mix = []

    for prs in prs_ray:
        py_mix = py_mix.condition(prs, temp)

        mfac = py_mix.mass_fract()
        vfac = py_mix.volm_fract()

        mfac_oil.append(mfac[0])
        mfac_wat.append(mfac[1])
        mfac_gas.append(mfac[2])

        vfac_oil.append(vfac[0])
        vfac_wat.append(vfac[1])
        vfac_gas.append(vfac[2])

        rho_mix.append(py_mix.rho_mix())

    pymix = {
        "mass_fracs": {"oil": mfac_oil, "wat": mfac_wat, "gas": mfac_gas},
        "volm_fracs": {"oil": vfac_oil, "wat": vfac_wat, "gas": vfac_gas},
        "rho_mix": rho_mix,
    }
    return pymix


@pytest.fixture(scope="module")
def hysys_resmix():
    hysys_path = Path(__file__).parents[1] / "data" / "hysys_resmix_peng_rob.json"
    with open(hysys_path) as json_file:
        return json.load(json_file)


@pytest.fixture(scope="module")
def python_resmix(hysys_resmix):
    return compute_resmix_data(
        hysys_resmix["pres_psig"],
        hysys_resmix["temp_degf"],
        hysys_resmix["watercut"],
        hysys_resmix["fgor"],
        hysys_resmix["oil_api"],
        hysys_resmix["pbub"],
        hysys_resmix["gas_sg"],
    )


def test_mass_fractions(hysys_resmix, python_resmix) -> None:
    name_frac = "mass_fracs"
    np.testing.assert_allclose(hysys_resmix[name_frac]["oil"], python_resmix[name_frac]["oil"], rtol=0.01)
    np.testing.assert_allclose(hysys_resmix[name_frac]["wat"], python_resmix[name_frac]["wat"], rtol=0.01)
    np.testing.assert_allclose(hysys_resmix[name_frac]["gas"], python_resmix[name_frac]["gas"], rtol=0.06)


def test_volm_fractions(hysys_resmix, python_resmix) -> None:
    name_frac = "volm_fracs"
    np.testing.assert_allclose(hysys_resmix[name_frac]["oil"], python_resmix[name_frac]["oil"], rtol=0.03)
    np.testing.assert_allclose(hysys_resmix[name_frac]["wat"], python_resmix[name_frac]["wat"], rtol=0.04)
    np.testing.assert_allclose(hysys_resmix[name_frac]["gas"], python_resmix[name_frac]["gas"], rtol=0.06)


def test_mixture_density(hysys_resmix, python_resmix) -> None:
    np.testing.assert_allclose(hysys_resmix["rho_mix"], python_resmix["rho_mix"], rtol=0.04)



"""Out Flow for an Oil Well

Performs the calculations that occur inside a vertical, inclined or horizontal wellbore.
Predominate equations are two phase flow correlations. The most common being Beggs and Brill
"""

import numpy as np

from woffl.flow import singlephase as sp
from woffl.flow import twophase as tp
from woffl.geometry import forms as fm
from woffl.geometry.pipe import PipeInPipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt import FormWater, ResMix


def homo_diff_press(
    pin: float,
    tin: float,
    hyd_dia: float,
    area: float,
    abs_ruff: float,
    length: float,
    height: float,
    qoil_std: float,
    prop: ResMix,
) -> tuple[float, float, float]:
    """Homogenous Differential Pressure

    Calculate Homogenous Differential Pressure Across Wellbore / Pipe.
    Hydraulic Diameter and Cross Sectional Area are defined separately to allow for annulus friction pressure
    drop calculations in reverse or forward ciruclating jet pumps.

    Args:
        pin (float): Inlet Pressure, psig
        tin (float): Inlet Temperature, deg F
        hyd_dia (float): Hydraulic Diameter of Pipe, inches
        area (float): Cross Sectional Area of Pipe, ft2
        abs_ruff (float): Absolute Roughness of Pipe, inches
        length (float): Length of Pipe, feet, Positive is with flow direction
        height (float): Height Difference of Pipe, feet, Positive is upwards
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix

    Returns:
        dp_stat (float): Static Differential Pressure, psid
        dp_fric (float): Friction Differential Pressure, psid
        nslh (float): No Slip Liquid Holdup, unitless
    """
    prop = prop.condition(pin, tin)
    qtot = sum(prop.insitu_volm_flow(qoil_std))
    rho_mix = prop.rho_mix()
    visc_mix = prop.visc_mix()
    nslh = prop.nslh()

    # area = sp.cross_area(inn_dia)
    vmix = sp.velocity(qtot, area)
    NRe = sp.reynolds(rho_mix, vmix, hyd_dia, visc_mix)
    rel_ruff = sp.relative_roughness(hyd_dia, abs_ruff)
    ff = sp.ffactor_darcy(NRe, rel_ruff)

    dp_fric = sp.diff_press_friction(ff, rho_mix, vmix, hyd_dia, length)
    dp_stat = sp.diff_press_static(rho_mix, height)
    return dp_stat, dp_fric, nslh


def beggs_diff_press(
    pin: float,
    tin: float,
    hyd_dia: float,
    area: float,
    abs_ruff: float,
    length: float,
    height: float,
    qoil_std: float,
    prop: ResMix,
) -> tuple[float, float, float]:
    """Beggs Differential Pressure

    Calculate Beggs Differential Pressure Across Wellbore / Pipe.
    Hydraulic Diameter and Cross Sectional Area are defined separately to allow for annulus friction pressure
    drop calculations in reverse or forward ciruclating jet pumps.

    Args:
        pin (float): Inlet Pressure, psig
        tin (float): Inlet Temperature, deg F
        hyd_dia (float): Hydraulic Diameter of Pipe, inches
        area (float): Cross Sectional Area of Pipe, ft2
        abs_ruff (float): Absolute Roughness of Pipe, inches
        length (float): Length of Pipe, feet, Positive is with flow direction
        height (float): Height Difference of Pipe, feet, Positive is upwards
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix

    Returns:
        dp_stat (float): Static Differential Pressure, psid
        dp_fric (float): Friction Differntial Pressure, psid
        slh (float): Slip Liquid Holdup, unitless
    """
    prop = prop.condition(pin, tin)
    qoil, qwat, qgas = prop.insitu_volm_flow(qoil_std)
    rho_liq, rho_gas = prop.rho_two()
    sig_liq = prop.tension
    nslh = prop.nslh()

    visc_mix = prop.visc_mix()
    rho_mix = prop.rho_mix()

    # area = sp.cross_area(inn_dia)
    vsl, vsg, vmix = tp.velocities(qoil, qwat, qgas, area)
    NFr = tp.froude(vmix, hyd_dia)
    NLv = tp.ros_nlv(vsl, rho_liq, sig_liq)  # ros liquid velocity number

    hpat, tran = tp.beggs_flow_pattern(nslh, NFr)
    incline = fm.horz_angle(length, height)
    slh = tp.beggs_holdup_inc(nslh, NFr, NLv, incline, hpat, tran)
    slh = min(slh, 1)  # liquid holdup never above one (before or after payne?)
    slh = tp.payne_correction(slh, incline)  # 1979 correction
    rho_slip = tp.density_slip(rho_liq, rho_gas, slh)
    dp_stat = tp.beggs_press_static(rho_slip, height)

    NRe = sp.reynolds(rho_mix, vmix, hyd_dia, visc_mix)
    rel_ruff = sp.relative_roughness(hyd_dia, abs_ruff)
    ff = sp.ffactor_darcy(NRe, rel_ruff)  # make this nomenclature consistent with beggs?
    yb = tp.beggs_yf(nslh, slh)  # NSLh, Lh
    sb = tp.beggs_sf(yb)
    fb = tp.beggs_ff(ff, sb)
    dp_fric = tp.beggs_press_friction(fb, rho_mix, vmix, hyd_dia, length)
    return dp_stat, dp_fric, slh


def production_top_down_press(
    ptop: float,
    ttop: float,
    qoil_std: float,
    prop: ResMix,
    wellbore: PipeInPipe,
    wellprof: WellProfile,
    flowpath: str = "tubing",
    model: str = "beggs",
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Production Top Down WellBore Pressure Calculation

    Uses the specificed model to calculate the pressure gradient in a wellbore, starting
    at the top and working down to inflow / jet pump node. This is the preferred method
    of calculation, as a negative pressure cannot be created during calculations.

    Args:
        ptop (float): Pressure at top node, psig
        ttop (float): Temperature at top node, deg F
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix
        wellbore (PipeInPipe): Piping geometry inside the wellbore, PipeInPipe
        wellprof (WellProfile): survey dimensions and location of jet pump, WellProfile
        flowpath (str): Where the flow is occuring, either tubing or annulus
        model (str): Specify which model to use, either homo or beggs

    Returns:
        md_seg (list): Measured depth of calculated pressure
        prs_ray (list): Calculated pressure along wellbore, psig
        slh_ray (list): Liquid Holdup along wellbore, unitless
    """
    flow_path_list = ["tubing", "annulus"]
    if flowpath not in flow_path_list:
        raise ValueError(f"{flowpath} not recognized, select from {flow_path_list}")
    if flowpath == "tubing":
        hyd_dia = wellbore.tube_hyd_dia
        area = wellbore.tube_area
        abs_ruff = wellbore.tube_abs_ruff
    else:
        hyd_dia = wellbore.ann_hyd_dia
        area = wellbore.ann_area
        abs_ruff = wellbore.ann_abs_ruff

    prs_ray = np.array([ptop])
    slh_ray = np.array([])
    md_seg, vd_seg = wellprof.outflow_spacing(100)  # space every 100'
    md_diff = np.diff(md_seg, n=1) * -1  # against flow
    vd_diff = np.diff(vd_seg, n=1) * -1  # going down piping
    n = 0
    for length, height in zip(md_diff, vd_diff):
        dp_stat, dp_fric, slh = beggs_diff_press(
            prs_ray[-1], ttop, hyd_dia, area, abs_ruff, length, height, qoil_std, prop
        )
        pdwn = prs_ray[-1] - dp_stat - dp_fric  # dp is subtracted
        prs_ray = np.append(prs_ray, pdwn)
        slh_ray = np.append(slh_ray, slh)
        n = n + 1
    # the no slip array is going to be one shorter than the md_seg and prs_ray...
    # i'm not sure if this is problem that I should "fix" later?
    return md_seg, prs_ray, slh_ray


def production_bottom_up_press(
    pbot: float,
    tbot: float,
    qoil_std: float,
    prop: ResMix,
    wellbore: PipeInPipe,
    wellprof: WellProfile,
    flowpath: str = "tubing",
    model: str = "beggs",
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """NOT TESTED: Bottom Up WellBore Pressure Calculation

    Uses the specificed model to calculate the pressure gradient in a wellbore, starting
    at the bottom inflow / jet pump node and working up to ther surface. Problem with this
    method is that a negative pressure can be calculated, causing an error.

    Args:
        pbot (float): Pressure at bottom node, psig
        tbot (float): Temperature at bottom node, deg F
        qoil_std (float): Oil Rate, STBOPD
        prop (ResMix): Properties of Fluid Mixture in Wellbore, ResMix
        wellbore (PipeInPipe): Piping geometry inside the wellbore, PipeInPipe
        wellprof (WellProfile): survey dimensions and location of jet pump, WellProfile
        flowpath (str): Where the flow is occuring, either tubing or annulus
        model (str): Specify which model to use, either homo or beggs

    Returns:
        md_seg (list): Measured depth of calculated pressure
        prs_ray (list): Calculated pressure along wellbore, psig
        slh_ray (list): Liquid Holdup along wellbore, unitless
    """
    flow_path_list = ["tubing", "annulus"]
    if flowpath not in flow_path_list:
        raise ValueError(f"{flowpath} not recognized, select from {flow_path_list}")
    if flowpath == "tubing":
        hyd_dia = wellbore.tube_hyd_dia
        area = wellbore.tube_area
        abs_ruff = wellbore.tube_abs_ruff
    else:
        hyd_dia = wellbore.ann_hyd_dia
        area = wellbore.ann_area
        abs_ruff = wellbore.ann_abs_ruff

    prs_ray = np.array([pbot])
    slh_ray = np.array([])
    md_seg, vd_seg = wellprof.outflow_spacing(100)  # space every 100'
    md_diff = np.diff(md_seg, n=1)  # with the flow
    vd_diff = np.diff(vd_seg, n=1)  # going up piping
    n = 0
    for length, height in zip(np.flip(md_diff), np.flip(vd_diff)):  # start at bottom
        dp_stat, dp_fric, slh = beggs_diff_press(
            prs_ray[-1], tbot, hyd_dia, area, abs_ruff, length, height, qoil_std, prop
        )
        pdwn = prs_ray[-1] - dp_stat - dp_fric  # dp is subtracted
        prs_ray = np.append(prs_ray, pdwn)
        slh_ray = np.append(slh_ray, slh)
        n = n + 1
    # the no slip array is going to be one shorter than the md_seg and prs_ray...
    # i'm not sure if this is problem that I should "fix" later?
    return md_seg, prs_ray, slh_ray


def powerfluid_top_down_friction(
    ptop: float,
    ttop: float,
    qwat_bpd: float,
    prop_pf: FormWater,  # update later to accept deadoil? need to integrate with resmix
    wellbore: PipeInPipe,
    wellprof: WellProfile,
    flowpath: str = "annulus",
) -> float:
    """Power Fluid Top Down Friction Calculation

    Used for calculating frictional pressure drop in the annlus (reverse circulating) or
    in the tubing (forward circulating) depennding on flow path. Prop is the
    power fluid properties, which is FormWater class, but really is just water.

    Args:
        ptop (float): Pressure at top node, psig
        ttop (float): Temperature at top node, deg F
        qwat_bpd (float): Water Rate, BWPD
        prop (FormWater): Poperties of Power Fluid
        wellbore (PipeInPipe): Piping geometry inside the wellbore, PipeInPipe
        wellprof (WellProfile): survey dimensions and location of jet pump, WellProfile
        flowpath (str): Where the flow is occuring, either "tubing" or "annulus"

    Returns:
        dp_fric (float): Friction Differntial Pressure, psid
    """
    flow_path_list = ["tubing", "annulus"]
    if flowpath not in flow_path_list:
        raise ValueError(f"{flowpath} not recognized, select from {flow_path_list}")
    if flowpath == "tubing":
        hyd_dia = wellbore.tube_hyd_dia
        area = wellbore.tube_area
        abs_ruff = wellbore.tube_abs_ruff
    else:
        hyd_dia = wellbore.ann_hyd_dia
        area = wellbore.ann_area
        abs_ruff = wellbore.ann_abs_ruff

    # height, positive is up?
    length = wellprof.jetpump_md  # distance down wellbore to jet pump

    prop_pf = prop_pf.condition(ptop, ttop)
    qwat_fts = sp.bpd_to_ft3s(qwat_bpd)
    vel = sp.velocity(qwat_fts, area)
    NRe = sp.reynolds(prop_pf.density, vel, hyd_dia, prop_pf.viscosity)
    rel_ruff = sp.relative_roughness(hyd_dia, abs_ruff)
    ff = sp.ffactor_darcy(NRe, rel_ruff)

    dp_fric = sp.diff_press_friction(ff, prop_pf.density, vel, hyd_dia, length)
    return dp_fric



"""Single Phase Flow Equations

Equations that are predominately used in single phase flows. These equations can
also be adapted and used in two phase flows, but require a simplification such as
no slip or homogenous mixing.
"""

import math


# unit conversion
def bpd_to_ft3s(q_bpd: float) -> float:
    """Convert liquid BPD to ft3/s

    Args:
        q_bpd (float): Volumetric Flow, BPD

    Returns:
        q_ft3s (float): Volumetric Flow, ft3/s
    """
    # 42 gal/bbl, 7.48052 gal/ft3, 24 hr/day, 60min/hour, 60sec/min
    q_ft3s = q_bpd * 42 / (24 * 60 * 60 * 7.48052)
    return q_ft3s


# unit conversion
def ft3s_to_bpd(q_ft3s: float) -> float:
    """Convert liquid ft3/s to BPD

    Args:
        q_ft3s (float): Volumetric Flow, ft3/s

    Returns:
        q_bpd (float): Volumetric Flow, BPD
    """
    # 42 gal/bbl, 7.48052 gal/ft3, 24 hr/day, 60min/hour, 60sec/min
    q_bpd = q_ft3s * (24 * 60 * 60 * 7.48052) / 42
    return q_bpd


def cross_area(dia: float) -> float:
    """Cross Sectional Area

    Calculate cross sectional area of piece of piping.

    Args:
        dia (float): Diameter, inches

    Returns:
        area (float): Area, ft2
    """
    dia = dia / 12  # ft
    area = math.pi * (dia**2) / 4  # ft2
    return area


def mom_to_psi(mom: float, area: float) -> float:
    """Convert Momentum across an area to lbf/in2

    Args:
        mom (float): Fluid Momentum, lbm*ft/s2
        area (float): Cross Sectional Area, ft2

    Return:
        p_mom (float): Equivalent Pressure of Momentum, psi
    """
    p_mom = mom / (area * 32.174 * 144)
    return p_mom


def velocity(flow: float, area: float) -> float:
    """Velocity Single Phase

    Args:
        flow (float): Volumetric Flow, ft3/s
        area (float): Pipe Cross Sectional Area, ft2

    Returns:
        vel (float): Flow Velocity, ft/s
    """
    vel = flow / area
    return vel


def momentum(rho: float, vel: float, area: float) -> float:
    """Fluid Momentum

    Args:
        vel (float): Velocity of the Fluid, ft/s
        area (float): Cross Sectional Area of the Flow, ft2
        rho (float): Density of the Fluid, lbm/ft3

    Returns:
        mom_fld (float): Fluid Momentum, lbm*ft/s2
    """
    return rho * vel**2 * area


def massflow(rho: float, vel: float, area: float) -> float:
    """Fluid Mass Flow

    Args:
        vel (float): Velocity of the Fluid, ft/s
        area (float): Cross Sectional Area of the Flow, ft2
        rho (float): Density of the Fluid, lbm/ft3

    Returns:
        mflow (float): Fluid Mass Flow, lbm/s
    """
    return rho * vel * area


def reynolds(rho: float, vel: float, dhyd: float, visc: float) -> float:
    """Reynolds Number

    Args:
        rho (float): Fluid density, lbm/ft3
        vel (float): Flow velocity, ft/s
        dhyd (float): Hydraulic Diameter, inches
        visc (float): Dynamic Viscosity, cP

    Returns:
        reynolds (float): Reynolds number, unitless
    """
    # convert inch to feet
    dhyd = dhyd / 12  # feet

    # convert cp to lbm/(ft*s)
    visc = visc / 1488.2  # lbm/(ft*s)

    reynolds = rho * vel * dhyd / visc
    return reynolds


def relative_roughness(dhyd: float, abs_ruff: float) -> float:
    """Relative Rougness of a Piece of Pipe

    Args:
        dhyd (float): Hydraulic Diameter, inches
        abs_ruff (float): Absolute Roughness, inches

    Returns:
        rel_ruff (float): Relative Roughness, e/D unitless
    """
    return abs_ruff / dhyd


def serghide(reynolds: float, rel_ruff: float) -> float:
    """Serghide Equation for Darcy-Weisbach Friction Factor

    Args:
        reynolds (float): Reynolds Number, Re unitless.
        rel_ruff (float): Relative Roughness, e/D unitless

    Returns:
        ff (float): Darcy-Weisbach Piping Friction Factor, unitless

    References:
        - Cranes Technical Paper No. 410 Equation 1-21
    """
    a = -2 * math.log10((rel_ruff / 3.7) + (12 / reynolds))
    b = -2 * math.log10((rel_ruff / 3.7) + (2.51 * a / reynolds))
    c = -2 * math.log10((rel_ruff / 3.7) + (2.51 * b / reynolds))
    ff = (a - (b - a) ** 2 / (c - 2 * b + a)) ** -2
    return ff


def ffactor_darcy(reynolds: float, rel_ruff: float) -> float:
    """Friction Factor Darcy Weisbach for Piping

    Args:
        reynolds (float): Reynolds Number, Re unitless
        rel_ruff (float): Relative Roughness, e/D unitless

    Returns:
        ff (float): Darcy-Weisbach Friction Factor of Piping, unitless
    """
    # laminar / transistional flow
    if reynolds < 4000:
        ff = 64 / reynolds

    else:
        ff = serghide(reynolds, rel_ruff)
    return ff


# dp_friction, friction_dp, fric_dp, or dp_fric
def diff_press_friction(ff: float, rho: float, vel: float, dhyd: float, length: float) -> float:
    """Frictional Differential Pressure in Piping

    Calculate the frictional pressure loss in a piping system.
    Negative is calculation is against flow, positive is with the flow.
    Matters if doing bottom down or top up calculations.

    Args:
        ff (float): Darcy friction factor of the pipe, unitless
        rho (float): Density Fluid, lbm/ft3
        vel (float): Velocity Fluid, ft/s
        dhyd (float): Hydraulic Diameter, inches
        length (float): Length of Pipe Segment, feet

    Returns:
        dp_fric (float): Frictional Differential Pressure, psi
    """
    g = 32.174  # 1 lbf equals 32.174 lbm*ft/s2
    dhyd = dhyd / 12  # feet
    dp_fric = ff * rho * vel**2 * length / (2 * dhyd * g)  # lbf/ft2
    dp_fric = dp_fric / 144  # lbf/in2
    return dp_fric


# dp_static, static_dp, or stat_dp or dp_stat
def diff_press_static(rho: float, height: float) -> float:
    """Static or Gravity Differential Pressure

    Incompressible Fluid Static Differential Pressure.
    Positive height is up, negative height is down.

    Args:
        rho (float): Fluid Density, lbm/ft3
        height (float): Fluid Vertical Height, feet

    Returns:
        dp_stat (float): Static Differential Pressure, psi
    """
    dp_stat = rho * height / 144  # psi, gravity cancels each other out with US Units
    return dp_stat



"""Batchpump Smoke Tests

Smoke tests for batch_run, process_results, and search_run.
Uses the Schrader preset well with a small jet pump grid to keep runtime short.
Asserts physically reasonable outputs — not exact values.
"""

import pytest

from woffl.geometry.jetpump import JetPump
from tests.asm_helper import make_well, jp_list


# ---- fixture ----

@pytest.fixture(scope="module")
def e41_batch():
    """Single well with batch_run and process_results completed."""
    well = make_well("MPE-41", qwf=246, pwf=1049, pres=1400, wc=0.894, fgor=600)
    well.batch_run(jp_list)
    well.process_results()
    return well


# ---- batch_run tests ----

class TestBatchRun:
    """Smoke tests for BatchPump.batch_run"""

    def test_returns_results(self, e41_batch):
        df = e41_batch.df
        assert len(df) == len(jp_list)

    def test_some_pumps_converge(self, e41_batch):
        df = e41_batch.df
        valid = df.dropna(subset=["qoil_std"])
        assert len(valid) > 0, "No pumps converged"

    def test_oil_rate_positive(self, e41_batch):
        df = e41_batch.df.dropna(subset=["qoil_std"])
        assert (df["qoil_std"] > 0).all()

    def test_suction_pressure_positive(self, e41_batch):
        df = e41_batch.df.dropna(subset=["psu_solv"])
        assert (df["psu_solv"] > 0).all()

    def test_water_rates_positive(self, e41_batch):
        df = e41_batch.df.dropna(subset=["totl_wat"])
        assert (df["totl_wat"] > 0).all()
        assert (df["lift_wat"] > 0).all()


# ---- process_results tests ----

class TestProcessResults:
    """Smoke tests for BatchPump.process_results"""

    def test_semi_finalists_exist(self, e41_batch):
        assert e41_batch.df["semi"].sum() > 0

    def test_gradients_finite(self, e41_batch):
        semi = e41_batch.df[e41_batch.df["semi"]]
        assert semi["motwr"].notna().all()
        assert semi["molwr"].notna().all()

    def test_curve_fit_coefficients(self, e41_batch):
        assert len(e41_batch.coeff_totl) == 3
        assert len(e41_batch.coeff_lift) == 3


# ---- search_run tests ----

class TestSearchRun:
    """Smoke tests for BatchPump.search_run"""

    def test_returns_catalog_pump(self):
        well = make_well("MPE-41", qwf=246, pwf=1049, pres=1400, wc=0.894, fgor=600)
        seed = JetPump("12", "B")
        df = well.search_run(seed, lift_cost=0.03)
        assert len(df) == 1
        assert df["nozzle"].iloc[0] != "opt"
        assert df["throat"].iloc[0] != "opt"

    def test_oil_rate_positive(self):
        well = make_well("MPE-41", qwf=246, pwf=1049, pres=1400, wc=0.894, fgor=600)
        seed = JetPump("12", "B")
        df = well.search_run(seed, lift_cost=0.03)
        assert df["qoil_std"].iloc[0] > 0

    def test_does_not_overwrite_batch_df(self):
        well = make_well("MPE-41", qwf=246, pwf=1049, pres=1400, wc=0.894, fgor=600)
        well.batch_run(jp_list)
        batch_len = len(well.df)
        seed = JetPump("12", "B")
        well.search_run(seed, lift_cost=0.03)
        assert len(well.df) == batch_len, "search_run overwrote self.df"



"""Batch Jet Pump Runs

Contains code that is used to run multiple pumps at once to understand the
current conditions. Code outputs a formatted list of python dictionaries that
can be converted to a Pandas Dataframe or equivalent for analysis.
"""

import os
from dataclasses import dataclass
from itertools import product

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.axes import Axes
from scipy.optimize import curve_fit, minimize

import woffl.assembly.solopump as so
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import PipeInPipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt import FormWater, ResMix


@dataclass(frozen=True)
class BatchResult:
    """Code is currently not used"""

    wellname: str
    nozzle: str
    throat: str
    choked: bool
    qoil_std: float
    fwat_bwpd: float
    mach_te: float
    total_water: float
    total_wc: float


class BatchPump:
    "Used for running different nozzle and throat combinations and selecting the correct one."

    def __init__(
        self,
        pwh: float,
        tsu: float,
        ppf_surf: float,
        wellbore: PipeInPipe,
        wellprof: WellProfile,
        ipr_su: InFlow,
        prop_su: ResMix,
        prop_pf: FormWater,
        jpump_direction: str = "reverse",
        wellname: str = "na",
    ) -> None:
        """Batch Pump Solver

        Used for iterating across a wide range of different pumps. An adjacent dataclass will
        be made for storing the results. Could add an optional wellname here if desired?

        Args:
            pwh (float): Pressure Wellhead, psig
            tsu (float): Temperature Suction, deg F
            ppf_surf (float): Pressure Power Fluid Surface, psig
            wellbore (PipeInPipe): Pipe Class of the Wellbore
            wellprof (WellProfile): Well Profile Class
            ipr_su (InFlow): Inflow Performance Class
            prop_su (ResMix): Reservoir Mixture Properties
            prop_pf (FormWater): Powerfluid Properties
            jpump_direction (str): Jet Pump Direction, "reverse" or "forward"
            wellname (str): A unique identifier of the wellname
        """
        self.pwh = pwh
        self.tsu = tsu
        self.ppf_surf = ppf_surf
        self.wellbore = wellbore
        self.wellprof = wellprof
        self.ipr_su = ipr_su
        self.prop_su = prop_su
        self.prop_pf = prop_pf
        self.direction = jpump_direction
        self.wellname = wellname

    def update_press(self, kind: str, psig: float) -> None:
        """Update Pressure

        Used to update different pressures instead of re-initializing everything. Components that
        can be updated include the well head, the power fluid or the reservoir pressure.

        Args:
            kind (str): Kind of Pressure to update. Either "wellhead", "powerfluid" or "reservoir"
            psig (float): Pressure to update with, psig
        """
        # chat gpt thinks using the reservoir method will fail, since you are calling ipr_su
        press_map = {"wellhead": "pwh", "powerfluid": "ppf_surf", "reservoir": "ipr_su.pres"}

        # Validate the 'kind' argument
        if kind not in press_map:
            valid_kind = ", ".join(press_map.keys())
            raise ValueError(f"Invalid value for 'kind': {kind}. Expected {valid_kind}.")

        attr_name = press_map[kind]
        setattr(self, attr_name, psig)

    @staticmethod
    def jetpump_list(
        nozzles: list[str],
        throats: list[str],
        knz: float = 0.01,
        ken: float = 0.03,
        kth: float = 0.3,
        kdi: float = 0.4,
    ) -> list[JetPump]:
        """Create a list of Jet Pumps

        Automatically generate a list of different types of jet pumps. The list is fed into
        the batch run to understand performance. Could move this into batch run if desired.

        Args:
            nozzles (list): List of Nozzle Strings
            throats (list): List of Throat Ratios
            knz (float): Nozzle Friction Factor
            ken (float): Enterance Friction Factor
            kth (float): Throat Friction Factor
            kdi (float): Diffuser Friction Factor

        Returns:
            jp_list (list): List of JetPumps
        """
        jp_list = []
        for nozzle, throat in product(nozzles, throats):
            jp_list.append(JetPump(nozzle, throat, knz, ken, kth, kdi))
        return jp_list

    def _run_core(self, jetpumps: list[JetPump], debug: bool = False) -> pd.DataFrame:
        """Core Jet Pump Solver Loop

        Run through multiple jet pumps and return results as a DataFrame
        without modifying self.df.

        Args:
            jetpumps (list): List of JetPumps
            debug (bool): True - Errors are Raised, False - Errors are Stored

        Returns:
            df (DataFrame): DataFrame of Jet Pump Results
        """
        results = []
        for jetpump in jetpumps:
            try:
                psu_solv, sonic_status, qoil_std, fwat_bpd, lwat_bpd, mach_te = so.jetpump_solver(
                    self.pwh,
                    self.tsu,
                    self.ppf_surf,
                    jetpump,
                    self.wellbore,
                    self.wellprof,
                    self.ipr_su,
                    self.prop_su,
                    self.prop_pf,
                    self.direction,
                )
                result = {
                    "nozzle": jetpump.noz_no,
                    "throat": jetpump.rat_ar,
                    "sonic_status": sonic_status,
                    "mach_te": mach_te,
                    "psu_solv": psu_solv,
                    "qoil_std": qoil_std,
                    "form_wat": fwat_bpd,
                    "lift_wat": lwat_bpd,
                    "totl_wat": fwat_bpd + lwat_bpd,
                    "form_wor": fwat_bpd / qoil_std,
                    "totl_wor": (fwat_bpd + lwat_bpd) / qoil_std,
                    "error": "na",
                }

            except Exception as exc:
                if debug:
                    print(f"Failed on nozzle {jetpump.noz_no} and throat: {jetpump.rat_ar}")
                    raise exc
                else:
                    result = {
                        "nozzle": jetpump.noz_no,
                        "throat": jetpump.rat_ar,
                        "sonic_status": False,
                        "mach_te": np.nan,
                        "psu_solv": np.nan,
                        "qoil_std": np.nan,
                        "form_wat": np.nan,
                        "lift_wat": np.nan,
                        "totl_wat": np.nan,
                        "form_wor": np.nan,
                        "totl_wor": np.nan,
                        "error": exc,
                    }
            results.append(result)
        return pd.DataFrame(results)

    def batch_run(self, jetpumps: list[JetPump], debug: bool = False) -> pd.DataFrame:
        """Batch Run of Jet Pumps

        Run through multiple different types of jet pumps. Results will be stored in
        a dataframe where the results can be graphed and selected for the optimal pump.
        The dataframe is added to the class as a variable for future inspection.

        Args:
            jetpumps (list): List of JetPumps
            debug (bool): True - Errors are Raised, False - Errors are Stored

        Returns:
            df (DataFrame): DataFrame of Jet Pump Results
        """
        self.df = self._run_core(jetpumps, debug)
        return self.df

    def search_run(self, seed: JetPump, lift_cost: float = 0.03, debug: bool = False) -> pd.DataFrame:
        """Search Run using Nelder-Mead

        Use Nelder-Mead to find the continuous nozzle and throat diameters that maximize
        oil rate for the given well conditions. The optimal continuous diameters are then
        snapped to the nearest catalog jet pump. The final result is the actual performance
        of that catalog pump, solved through the full wellbore physics.

        The lift_cost penalizes lift water (power fluid) usage. It represents the oil
        production cost of each barrel of lift water, in bbl oil / bbl lift water. For
        example, a lift_cost of 0.03 means 3 bbls of oil per 100 bbls of lift water.
        A higher value favors smaller pumps that use less power fluid. A value of 0.0
        maximizes oil with no regard for water usage.

        Args:
            seed (JetPump): Starting jet pump to seed the optimizer
            lift_cost (float): Lift water penalty, bbl oil / bbl lift water
            debug (bool): True - Errors are Raised, False - Errors are Stored

        Returns:
            df (DataFrame): DataFrame with the catalog pump result and continuous optimum
        """

        def objective(x):
            dnz, dth = x
            if dth <= dnz:
                return 1e10
            jp = continuous_jetpump(dnz, dth, seed.knz, seed.ken, seed.kth, seed.kdi)
            try:
                _, _, qoil, _, lwat, _ = so.jetpump_solver(
                    self.pwh,
                    self.tsu,
                    self.ppf_surf,
                    jp,
                    self.wellbore,
                    self.wellprof,
                    self.ipr_su,
                    self.prop_su,
                    self.prop_pf,
                    self.direction,
                )
                return -(qoil - lift_cost * lwat)
            except Exception:
                return 1e10

        x0 = [seed.dnz, seed.dth]
        bounds = [
            (JetPump.nozzle_dia[0], JetPump.nozzle_dia[-1]),
            (JetPump.throat_dia[0], JetPump.throat_dia[-1]),
        ]

        result = minimize(objective, x0, method="Nelder-Mead", bounds=bounds)
        dnz_opt, dth_opt = result.x

        # snap to the nearest catalog jet pump and run the actual physics
        best_jp = snap_to_catalog(dnz_opt, dth_opt, seed.knz, seed.ken, seed.kth, seed.kdi)

        df = self._run_core([best_jp], debug)

        # store the continuous optimum for reference
        df["dnz_opt"] = dnz_opt
        df["dth_opt"] = dth_opt

        return df

    def process_results(self) -> pd.DataFrame:
        """Process Results

        Identify the semi-finalist jet pumps, calculate numerical gradients and curve fits.
        The semi-fialist jet pump means no other jet pump makes more oil for less water.
        The gradients are MOTWR, stands for marginal oil total water ratio and MOLWR, stands
        for marginal oil lift water ratio. Calculate curve fit coefficients for the jet
        pump lift and total water vs oil.

        Args:
            self

        Returns:
            df (DataFrame): Adds semi (bool) column, motwr, molwr.
        """
        semi_mask = batch_results_mask(self.df["qoil_std"], self.df["totl_wat"], self.df["nozzle"])
        self.df["semi"] = semi_mask

        semi_df = self.df[self.df["semi"]].copy()
        if semi_df.empty:
            raise ValueError(f"No Semi-Finalist Jet Pumps on {self.wellname}")

        semi_df = semi_df.sort_values(by="qoil_std", ascending=True)

        qoil_semi = semi_df["qoil_std"].to_numpy()
        twat_semi = semi_df["totl_wat"].to_numpy()
        lwat_semi = semi_df["lift_wat"].to_numpy()

        semi_df["motwr"] = gradient_back(qoil_semi, twat_semi)
        semi_df["molwr"] = gradient_back(qoil_semi, lwat_semi)

        self.df = self.df.merge(semi_df[["motwr", "molwr"]], left_index=True, right_index=True, how="left")

        self.coeff_totl = batch_curve_fit(qoil_semi, twat_semi, origin=False)
        self.coeff_lift = batch_curve_fit(qoil_semi, lwat_semi, origin=False)

        return self.df

    def plot_data(
        self, water: str, curve: bool = False, ax: Axes | None = None, fig_path: str | os.PathLike | None = None
    ) -> None:
        """Plot Data

        Plot the results from the jet pump batch run to visualize the performance
        of each jet pump. The figure path allows the figure to be locally saved if passed.

        Args:
            water (str): "lift" or "total" depending on the desired x axis
            curve (bool): Show the curve fit or not
            ax (Axes): Matplotlib axes
            fig_path (Path): Path to optional output file, saves files to be viewed later
        """
        water = validate_water(water)

        # Determine the correct water data and coefficients
        qwat_bpd = self.df["lift_wat"] if water == "lift" else self.df["totl_wat"]
        coeff = self.coeff_lift if water == "lift" else self.coeff_totl  # type: ignore
        coeff = coeff if curve else None

        hold = 1  # define the variable with anything, need to trade
        if ax is None:
            hold = None  # transfer None value to something not used
            fig, ax = plt.subplots(figsize=(5.5, 4))

        # calling the function to plot the data
        batch_plot_data(
            qoil_std=self.df["qoil_std"],
            qwat_bpd=qwat_bpd,
            water=water,
            nozzles=self.df["nozzle"],
            throats=self.df["throat"],
            semi=self.df["semi"],
            wellname=self.wellname,
            coeff=coeff,
            ax=ax,  # type: ignore
        )

        if hold is None:
            if fig_path is not None:
                plt.savefig(fig_path, dpi=300)
            else:
                plt.show()

    def plot_derv(self, water: str, fig_path: str | os.PathLike | None = None) -> None:
        """Plot Derivative

        Plot the derivative results from the jet pump batch run to visualize how
        well of a match occured between the data and curve for each jet pump

        Args:
            water (str): "lift" or "total" depending on the desired x axis
            fig_path (Path): Path to optional output file, saves files to be viewed later
        """
        # Validate the 'water' argument
        water = validate_water(water)
        df_semi = self.df[self.df["semi"]]  # filter out the bad parts

        # Determine the correct water data and coefficients
        marg_semi = df_semi["molwr"] if water == "lift" else df_semi["motwr"]
        qwat_semi = df_semi["lift_wat"] if water == "lift" else df_semi["totl_wat"]
        coeff = self.coeff_lift if water == "lift" else self.coeff_totl

        fig, ax = plt.subplots(figsize=(5.5, 4))

        # calling the function to plot the derivatives
        batch_plot_derv_base(
            marg_filt=marg_semi,
            qwat_filt=qwat_semi,
            nozz_filt=df_semi["nozzle"],
            thrt_filt=df_semi["throat"],
            wellname=self.wellname,
            coeff=coeff,
            ax=ax,
            mcolor="blue",
            network=False,
        )

        ax.set_xlabel(f"{water.capitalize()} Water Rate, BWPD")
        ax.set_ylabel("Marginal Oil Water Rate, Oil BBL / Water BBL")
        ax.title.set_text(f"{self.wellname} Jet Pump Performance")
        ax.legend()

        if fig_path is not None:
            plt.savefig(fig_path, dpi=300)
        else:
            plt.show()


def validate_water(water: str) -> str:
    """Validate Type of Water String

    Checks that the string passed into a method or arguement fits the required description.
    This is used when the water type wants to be defined as lift or total

    Args:
        water (str): "lift" or "total" depending on the desired x axis

    Returns:
        water (str): Properly formatted as either "lift" or "total"
    """
    # Validate the 'water' argument
    if water not in {"lift", "total", "totl"}:
        raise ValueError(f"Invalid value for 'water': {water}. Expected 'lift', 'total', or 'totl'.")

    # Standardize "totl" to "total"
    if water == "totl":
        water = "total"
    return water


def batch_results_mask(
    qoil_std: np.ndarray | pd.Series,
    qwat_tot: np.ndarray | pd.Series,
    nozzles: np.ndarray | pd.Series,
) -> np.ndarray:
    """Batch Results Mask

    Create a mask of booleans from batch results. The initial filter is selecting the throat with the highest
    oil rate for each nozzle size. Any points where the oil rate is lower for a higher amount of
    water are then next removed. The filtered points can be passed to a function to calculate the gradient.

    Args:
        qoil_std (np.ndarray | pd.Series): Oil Prod. Rate, BOPD
        qwat_tot (np.ndarray | pd.Series): Total Water Rate, BWPD
        nozzles (np.ndarray | pd.Series): List of the Nozzles

    Returns:
        np.ndarray: True is a point to calc gradient, False means a point exists with better oil and less water
    """
    # convert into numpy arrays
    qoil_std = np.asarray(qoil_std)
    qwat_tot = np.asarray(qwat_tot)
    nozzles = np.asarray(nozzles)

    mask = np.zeros(len(qoil_std), dtype=bool)

    # start by picking the highest oil rate for each nozzle
    unique_nozzles = np.unique(nozzles)  # unique nozzles in the list
    for noz in unique_nozzles:
        noz_idxs = np.where(nozzles == noz)[0]  # indicies where the nozzle is a specific nozzle

        if np.all(np.isnan(qoil_std[noz_idxs])):  # skip the nozzle if they are all nan
            continue

        best_idx = noz_idxs[np.nanargmax(qoil_std[noz_idxs])]
        mask[best_idx] = True

    # compare the points to themselves to look for where oil is higher for less water
    for idx in np.where(mask)[0]:
        higher_wat_mask = qwat_tot < qwat_tot[idx]
        lower_oil_mask = qoil_std > qoil_std[idx]
        if np.any(higher_wat_mask & lower_oil_mask):
            mask[idx] = False

    return mask


def gradient_back(oil_rate: np.ndarray, water_rate: np.ndarray) -> list:
    """Gradient Calculations Feed Backwards

    Completes a numerical gradient calculation between jet pumps. Adds a zero on
    the oil and water arrays that are passed to it. These ensures points exists for
    doing a reverse gradient calculation.

    Args:
        qoil_filt (list): Filtered Oil Prod. Rate, BOPD
        qwat_filt (list): Filtered Total Water Rate, BWPD

    Returns:
        gradient (list): Gradient of Oil-Rate / Water-Rate, bbl/bbl
    """
    oil_rate = np.append(0, oil_rate)  # add a zero to the start
    water_rate = np.append(0, water_rate)  # add a zero to the start

    if len(oil_rate) != len(water_rate):
        raise ValueError("Oil and Water Arrays Must be the same length")

    grad = []
    for i in range(len(oil_rate)):
        if i != 0:  # skip the first value of 0,0, since you aren't returning that value
            grad.append((oil_rate[i] - oil_rate[i - 1]) / (water_rate[i] - water_rate[i - 1]))
        else:
            pass
    return grad


def batch_plot_data(
    qoil_std: pd.Series,
    qwat_bpd: pd.Series,
    water: str,
    nozzles: pd.Series,
    throats: pd.Series,
    semi: pd.Series,
    wellname: str,
    coeff: tuple[float, float, float] | None,
    ax: plt.Axes,  # type: ignore
) -> None:
    """Batch Plot Data

    Create a plot to view the results from the batch run.
    Add an additional argument that would be the coefficients for
    the curve fit of the upper portion of the data.

    Args:
        qoil_std (pd.Series): Oil Prod. Rate, BOPD
        qwat_bpd (pd.Series): Water Rate, "Lift" or "Total", BPD
        water (str): Water Type, "Lift" or "Total"
        nozzles (pd.Series): Nozzle Numbers
        throats (pd.Series): Throat Ratios
        semi (pd.Series): Pandas Series of if the jet pump is a semi finalist or not
        wellname (str): Wellname String
        coeff (tuple): Exponential Curve Fit Coefficients, A, B, C
        ax (plt.Axes): Matplotlib Axes
    """
    jp_names = [noz + thr for noz, thr in zip(nozzles, throats)]  # create a list of all the jetpump names

    # plot semi-finalist
    ax.plot(qwat_bpd[semi], qoil_std[semi], marker="o", linestyle="", color="r", label="Semi-Final")

    # plot non-semi finalist
    ax.plot(qwat_bpd[~semi], qoil_std[~semi], marker="o", linestyle="", color="b", label="Eliminate")

    for qoil, qwat, jp in zip(qoil_std, qwat_bpd, jp_names):
        if not pd.isna(qoil):
            ax.annotate(jp, xy=(qwat, qoil), xycoords="data", xytext=(1.5, 1.5), textcoords="offset points")

    # show the curve fit data
    if coeff is not None:
        a, b, c = coeff  # parse out the coefficients for easier understanding
        fit_water = np.linspace(0, qwat_bpd.max(), 1000)
        fit_oil = [exp_model(wat, a, b, c) for wat in fit_water]
        ax.plot(fit_water, fit_oil, color="red", linestyle="--", label="Exp. Fit")

    ax.set_xlabel(f"{water.capitalize()} Water Rate, BWPD")
    ax.set_ylabel("Produced Oil Rate, BOPD")
    ax.title.set_text(f"{wellname} Jet Pump Performance")
    ax.legend()

    return None


def batch_plot_derv_base(
    marg_filt: pd.Series,
    qwat_filt: pd.Series,
    nozz_filt: pd.Series,
    thrt_filt: pd.Series,
    wellname: str,
    coeff: tuple[float, float, float],
    ax: Axes,
    mcolor: str | mcolors.Colormap,
    network: bool = False,
) -> None:
    """Batch Plot Derivative Base

    Create a plot showing the derivative / marginal oil results.
    Visualize how good the curve fit actually is for the data.
    Data needs to be filtered prior to passing to this function.

    Args:
        marg_filt (pd.Series): Filtered Marginal Oil - Water Ratio Rate, BOWPD
        qwat_filt (pd.Series): Filtered Water Rate, "Lift" or "Total", BPD
        nozz_filt (pd.Series): Filtered Nozzle Numbers
        thrt_filt (pd.Series): Filtered Throat Ratios
        wellname (str): Wellname String
        coeff (tuple): Exponential Curve Fit Coefficients, A, B, C
        ax (plt.Axes): Matplotlib Axes
        mcolor (str): Matplotlib Color
        network (bool): Will this be used in a network plot or not
    """
    if network:
        label1 = wellname
        label2 = "_hidden"
    else:
        label1 = "Numerical"
        label2 = "Analytical"

    ax.plot(qwat_filt, marg_filt, marker="o", linestyle="", color=mcolor, label=label1)

    jp_filt = [noz + thr for noz, thr in zip(nozz_filt, thrt_filt)]  # create a list of all the jetpump names
    for marg, qwat, jp in zip(marg_filt, qwat_filt, jp_filt):
        ax.annotate(jp, xy=(qwat, marg), xycoords="data", xytext=(1.5, 1.5), textcoords="offset points")

    a, b, c = coeff  # parse out the coefficients for easier understanding
    fit_water = np.linspace(0, qwat_filt.max(), 1000)
    fit_grad = [exp_deriv(wat, b, c) for wat in fit_water]
    ax.plot(fit_water, fit_grad, color=mcolor, linestyle="--", label=label2)

    return None


def continuous_jetpump(
    dnz: float,
    dth: float,
    knz: float = 0.01,
    ken: float = 0.03,
    kth: float = 0.3,
    kdi: float = 0.3,
) -> JetPump:
    """Create a Jet Pump with Arbitrary Continuous Diameters

    Bypass the catalog lookup in JetPump.__init__ and set nozzle and throat
    diameters directly. Used by the Nelder-Mead optimizer to evaluate
    non-catalog pump geometries during the search.

    Args:
        dnz (float): Nozzle Diameter, inches
        dth (float): Throat Diameter, inches
        knz (float): Nozzle Friction Factor, unitless
        ken (float): Enterance Friction Factor, unitless
        kth (float): Throat Friction Factor, unitless
        kdi (float): Diffuser Friction Factor, unitless

    Returns:
        jp (JetPump): Jet Pump with the specified diameters
    """
    jp = object.__new__(JetPump)
    jp.noz_no = "opt"
    jp.rat_ar = "opt"
    jp.knz = knz
    jp.ken = ken
    jp.kth = kth
    jp.kdi = kdi
    jp.dnz = dnz
    jp.dth = dth
    return jp


def snap_to_catalog(
    dnz_opt: float,
    dth_opt: float,
    knz: float = 0.01,
    ken: float = 0.03,
    kth: float = 0.3,
    kdi: float = 0.3,
) -> JetPump:
    """Snap Continuous Diameters to Nearest Catalog Jet Pump

    Find the valid nozzle and area ratio combination whose diameters are
    closest (Euclidean distance) to the continuous optimum. Searches all
    valid nozzle/area_ratio pairs in the Champion X catalog.

    Args:
        dnz_opt (float): Optimal Nozzle Diameter, inches
        dth_opt (float): Optimal Throat Diameter, inches
        knz (float): Nozzle Friction Factor, unitless
        ken (float): Enterance Friction Factor, unitless
        kth (float): Throat Friction Factor, unitless
        kdi (float): Diffuser Friction Factor, unitless

    Returns:
        jp (JetPump): Nearest valid catalog Jet Pump
    """
    noz_dia = np.array(JetPump.nozzle_dia)
    thr_dia = np.array(JetPump.throat_dia)

    # snap nozzle first, then find best valid throat ratio
    noz_idx = np.argmin((noz_dia - dnz_opt) ** 2)

    best_letter = min(
        JetPump.area_code,
        key=lambda j: (
            (thr_dia[noz_idx + JetPump.area_code[j]] - dth_opt) ** 2
            if 0 <= noz_idx + JetPump.area_code[j] < len(thr_dia)
            else np.inf
        ),
    )

    return JetPump(str(noz_idx + 1), best_letter, knz, ken, kth, kdi)


def exp_model(x: float, a: float, b: float, c: float) -> float:
    """Exponential Curve Fit

    Args:
        x (float): Water Rate, bwpd
        a (float): Asymptote of the Curve
        b (float): Constant
        c (float): Constant

    Returns
        y (float): Oil Rate, bopd
    """
    return a - b * np.exp(-c * x)


def exp_deriv(x: float, b: float, c: float) -> float:
    """Derivative of Exponential Curve Fit

    Args:
        x (float): Water Rate, bwpd
        b (float): Constant
        c (float): Constant

    Returns
        s (float): Marginal Oil - Water Ratio, bbl/bbl
    """
    return c * b * np.exp(-c * x)


def rev_exp_deriv(s: float, b: float, c: float) -> float:
    """Reverse Derivative of Exponential Curve Fit

    Args:
        s (float): Marginal Oil - Water Ratio, bbl/bbl
        b (float): Constant
        c (float): Constant

    Returns
        x (float): Water Rate, bwpd
    """
    if s == 0:
        s = 0.00001
    x = -1 / c * np.log(s / (c * b))
    x = max(x, 0)  # make sure s doesn't drop below zero
    return x


def batch_curve_fit(qoil_filt: np.ndarray, qwat_filt: np.ndarray, origin: bool = True) -> tuple[float, float, float]:
    """Batch Curve Fit

    Curve fit the filtered datapoints from the Batch Results

    Args:
        qoil_filt (list): Filtered Oil Array, bopd
        qwat_filt (list): Filtered Water Array, bwpd
        origin (bool): Add point to encourage intercepting at (0,0)

    Returns:
        coeff (float): a, b and c coefficients for curve fit
    """
    # add a point at 0,0 to force intercepting origin
    if origin:
        qoil_filt = np.append(qoil_filt, 0.0)
        qwat_filt = np.append(qwat_filt, 0.0)

    initial_guesses = [max(qoil_filt), max(qoil_filt), 0.001]
    coeff, _ = curve_fit(exp_model, qwat_filt, qoil_filt, p0=initial_guesses)
    return coeff



import math

import pytest

from woffl.geometry.jetpump import JetPump


class TestJetPumpInit:
    """Test that JetPump correctly looks up catalog nozzle and throat sizes."""

    def test_known_pump_12c(self):
        """12C is a common pump — verify nozzle and throat from catalog."""
        jp = JetPump("12", "C")
        assert jp.dnz == 0.2099  # nozzle_dia[11]
        assert jp.dth == 0.4335  # throat_dia[13]
        assert jp.noz_no == "12"
        assert jp.rat_ar == "C"

    def test_known_pump_9e(self):
        """9E — small nozzle, large throat ratio."""
        jp = JetPump("9", "E")
        assert jp.dnz == 0.1458  # nozzle_dia[8]
        assert jp.dth == 0.3841  # throat_dia[12]

    def test_known_pump_5a(self):
        """5A — nozzle_idx=4, throat_idx=4+0=4."""
        jp = JetPump("5", "A")
        assert jp.dnz == 0.0903  # nozzle_dia[4]
        assert jp.dth == 0.1458  # throat_dia[4]

    def test_known_pump_7x(self):
        """7X — area_code X is -1, so throat_idx = 6 + (-1) = 5."""
        jp = JetPump("7", "X")
        assert jp.dnz == 0.1145  # nozzle_dia[6]
        assert jp.dth == 0.1643  # throat_dia[5]

    def test_default_loss_coefficients(self):
        jp = JetPump("10", "B")
        assert jp.knz == 0.01
        assert jp.ken == 0.03
        assert jp.kth == 0.3
        assert jp.kdi == 0.3

    def test_custom_loss_coefficients(self):
        jp = JetPump("10", "B", knz=0.02, ken=0.05, kth=0.4, kdi=0.5)
        assert jp.knz == 0.02
        assert jp.ken == 0.05
        assert jp.kth == 0.4
        assert jp.kdi == 0.5


class TestJetPumpErrors:
    """Test that bad nozzle/throat combos raise ValueError."""

    def test_nozzle_too_large(self):
        with pytest.raises(ValueError, match="Nozzle size"):
            JetPump("25", "A")

    def test_nozzle_zero(self):
        with pytest.raises(ValueError, match="out of range"):
            JetPump("0", "A")

    def test_nozzle_negative(self):
        with pytest.raises(ValueError, match="out of range"):
            JetPump("-5", "A")

    def test_nozzle_non_numeric(self):
        with pytest.raises(ValueError):
            JetPump("abc", "A")

    def test_invalid_area_ratio(self):
        with pytest.raises(ValueError, match="not recognized"):
            JetPump("10", "Z")

    def test_throat_index_out_of_range(self):
        """20E — nozzle exists (index 19) but throat_idx = 19+4 = 23, out of range."""
        with pytest.raises(ValueError, match="Nozzle throat combo"):
            JetPump("20", "E")


class TestJetPumpProperties:
    """Test computed areas from catalog diameters."""

    def test_nozzle_area(self):
        jp = JetPump("12", "C")
        expected = math.pi * (0.2099**2) / 4 / 144
        assert jp.anz == pytest.approx(expected)

    def test_throat_area(self):
        jp = JetPump("12", "C")
        expected = math.pi * (0.4335**2) / 4 / 144
        assert jp.ath == pytest.approx(expected)

    def test_throat_entry_area(self):
        """ate = ath - anz, the annular area between nozzle OD and throat ID."""
        jp = JetPump("12", "C")
        assert jp.ate == pytest.approx(jp.ath - jp.anz)

    def test_throat_larger_than_nozzle(self):
        """For any valid pump, throat area must exceed nozzle area."""
        jp = JetPump("10", "A")
        assert jp.ath > jp.anz
        assert jp.ate > 0



"""Solopump Smoke Tests

Smoke tests for the single-well, single-pump solver (solopump.jetpump_solver).
Tests pump sizing response, power fluid pressure sensitivity, circulation
direction, and wellbore geometry effects.
"""

import woffl.assembly.solopump as so
from woffl.flow.inflow import InFlow
from woffl.geometry.jetpump import JetPump
from woffl.geometry.pipe import Pipe, PipeInPipe
from woffl.geometry.wellprofile import WellProfile
from woffl.pvt.blackoil import BlackOil
from woffl.pvt.formgas import FormGas
from woffl.pvt.formwat import FormWater
from woffl.pvt.resmix import ResMix

# --- E-41 baseline setup ---

pwh = 210  # psig
tsu = 80  # deg F
ppf_surf = 3168  # psig

tubing = Pipe(out_dia=4.5, thick=0.5)
casing = Pipe(out_dia=6.875, thick=0.5)
wbore = PipeInPipe(inn_pipe=tubing, out_pipe=casing)
profile = WellProfile.schrader()

mpu_oil = BlackOil.schrader()
mpu_wat = FormWater.schrader()
mpu_gas = FormGas.schrader()

ipr = InFlow(qwf=246, pwf=1049, pres=1400)
res = ResMix(wc=0.894, fgor=600, oil=mpu_oil, wat=mpu_wat, gas=mpu_gas)


def _solve(jpump, pwh=pwh, ppf=ppf_surf, wellbore=wbore, direction="reverse"):
    """Helper to call jetpump_solver with E-41 defaults, returns (qoil, lwat)."""
    psu, sonic, qoil, fwat, lwat, mach = so.jetpump_solver(
        pwh, tsu, ppf, jpump, wellbore, profile, ipr, res, mpu_wat, direction
    )
    return qoil, lwat


class TestPumpSizing:
    """Larger semi-finalist pump should produce more oil.

    Pumps selected from E-41 batch semi-finalists:
    11D (~193 bopd) and 13C (~225 bopd).
    """

    def test_larger_pump_more_oil(self):
        qoil_small, _ = _solve(JetPump("11", "D"))
        qoil_large, _ = _solve(JetPump("13", "C"))
        assert qoil_large > qoil_small

    def test_larger_pump_more_water(self):
        _, lwat_small = _solve(JetPump("11", "D"))
        _, lwat_large = _solve(JetPump("13", "C"))
        assert lwat_large > lwat_small


class TestPowerFluidPressure:
    """Higher power fluid pressure delivers more energy to the nozzle."""

    def test_higher_ppf_more_oil(self):
        pump = JetPump("12", "C")
        qoil_low, _ = _solve(pump, ppf=2500)
        qoil_high, _ = _solve(pump, ppf=3168)
        assert qoil_high > qoil_low


class TestCirculationDirection:
    """Reverse vs forward with similar tubing and annular flow areas.

    Tubing and casing were sized so the tubing cross-sectional area and the
    annular cross-sectional area are nearly equal, isolating the effect of
    flow path geometry (circular vs annular) on friction:

        Tubing: 4.5" OD, 0.50" wall -> ID=3.50", area=9.62 in2
        Annulus: 6.25" OD casing, 0.30" wall -> ann area=9.17 in2 (ratio 1.05)

    Expected: with similar areas, reverse circulation (production up tubing)
    would have slightly less oil than forward (production up annulus) because
    the tubing imposes more friction on the multiphase production stream.

    Actual: reverse produced significantly more oil (217 vs 143 bopd). Likely
    caused by Beggs & Brill performing poorly in annular geometry — the
    hydraulic diameter correction for an annulus may overpredict friction.
    Test asserts the two directions produce different results, without
    asserting which is higher, until the annular multiphase model is better
    understood.
    """

    def test_reverse_vs_forward(self):
        tube_dir = Pipe(out_dia=4.5, thick=0.50)
        case_dir = Pipe(out_dia=6.25, thick=0.30)
        wbore_dir = PipeInPipe(inn_pipe=tube_dir, out_pipe=case_dir)

        pump = JetPump("12", "C")
        qoil_rev, _ = _solve(pump, wellbore=wbore_dir, direction="reverse")
        qoil_fwd, _ = _solve(pump, wellbore=wbore_dir, direction="forward")

        # with similar areas, expect a difference — assert they're not equal
        # direction of difference TBD, log both for inspection
        print(f"\nReverse: {qoil_rev:.1f} bopd, Forward: {qoil_fwd:.1f} bopd")
        assert qoil_rev != qoil_fwd


class TestWellboreGeometry:
    """Smaller wellbore components increase friction and reduce oil."""

    def test_smaller_casing_less_oil(self):
        """Same tubing, smaller casing -> more annular friction -> less oil."""
        pump = JetPump("12", "C")

        big_casing = Pipe(out_dia=6.875, thick=0.5)
        small_casing = Pipe(out_dia=5.5, thick=0.25)
        wbore_big = PipeInPipe(inn_pipe=tubing, out_pipe=big_casing)
        wbore_small = PipeInPipe(inn_pipe=tubing, out_pipe=small_casing)

        qoil_big, _ = _solve(pump, wellbore=wbore_big)
        qoil_small, _ = _solve(pump, wellbore=wbore_small)
        assert qoil_big > qoil_small

    def test_smaller_tubing_less_oil(self):
        """Big dummy casing (9" OD) so annular friction is negligible.
        Smaller tubing -> more production friction -> less oil."""
        pump = JetPump("12", "C")
        dummy_casing = Pipe(out_dia=9.0, thick=0.25)

        big_tubing = Pipe(out_dia=4.5, thick=0.5)
        small_tubing = Pipe(out_dia=3.5, thick=0.271)
        wbore_big = PipeInPipe(inn_pipe=big_tubing, out_pipe=dummy_casing)
        wbore_small = PipeInPipe(inn_pipe=small_tubing, out_pipe=dummy_casing)

        qoil_big, _ = _solve(pump, wellbore=wbore_big)
        qoil_small, _ = _solve(pump, wellbore=wbore_small)
        assert qoil_big > qoil_small



import json
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
from scipy import optimize


class WellProfile:
    """Well Profile Class

    Create a wellprofile, which is the subsurface geometry of the measured depth versus vertical depth.
    Can be used to interpolate values for understanding how measured depth relates to vertical depth.
    """

    def __init__(self, md_list: list | np.ndarray, vd_list: list | np.ndarray, jetpump_md: float) -> None:
        """Create a Well Profile

        Args:
            md_list (list): List of measured depths
            vd_list (list): List of vertical depths
            jetpump_md (float): Measured depth of Jet Pump, feet

        Returns:
            Self
        """
        if len(md_list) != len(vd_list):
            raise ValueError("Lists for Measured Depth and Vertical Depth need to be the same length")

        if max(md_list) < max(vd_list):
            raise ValueError("Measured Depth needs to extend farther than Vertical Depth")

        if jetpump_md > max(md_list):
            raise ValueError("Jet pump not inside well profile measured depth")

        md_ray, vd_ray = sort_profile(np.array(md_list), np.array(vd_list))

        self.md_ray = md_ray
        self.vd_ray = vd_ray
        self.hd_ray = self._horz_dist(self.md_ray, self.vd_ray)
        self.jetpump_md = jetpump_md

        # here is the real question, do I even need the raw data? just filter the data
        # at the __init__ and be done? Is the raw data just more stuff to wade through?
        self.hd_fit, self.vd_fit, self.md_fit = self.filter()  # run the method

    def __repr__(self):
        final_md = round(self.md_ray[-1], 0)
        final_vd = round(self.vd_ray[-1], 0)

        return f"Profile is {final_md} ft. long and {final_vd} ft. deep"

    def vd_interp(self, md_dpth: float) -> float:
        """Vertical Depth Interpolation

        Args:
            md_dpth (float): Measured Depth, feet

        Returns:
            vd_dpth (float): Vertical Depth, feet
        """
        return self._depth_interp(md_dpth, self.md_ray, self.vd_ray)

    def hd_interp(self, md_dpth: float) -> float:
        """Horizontal Distance Interpolation

        Args:
            md_dpth (float): Measured Depth, feet

        Returns:
            hd_dist (float): Horizontal Distance, feet
        """
        return self._depth_interp(md_dpth, self.md_ray, self.hd_ray)

    def md_interp(self, vd_dpth: float) -> float:
        """Measured Depth Interpolation

        Args:
            vd_dpth (float): Vertical Depth, feet

        Returns:
            md_dpth (float): Measured Depth, feet
        """
        return self._depth_interp(vd_dpth, self.vd_ray, self.md_ray)

    @property
    def jetpump_vd(self) -> float:
        """Jet Pump True Vertical Depth, Feet"""
        jp_vd = self.vd_interp(self.jetpump_md)
        return jp_vd

    @property
    def jetpump_hd(self) -> float:
        """Jet Pump Horizontal Distance, Feet"""
        jp_hd = self.hd_interp(self.jetpump_md)
        return jp_hd

    # make a plot of the filtered data on top of the raw data
    def plot_raw(self) -> None:
        """Plot the Raw Profile Data"""

        self._profileplot(self.hd_ray, self.vd_ray, self.md_ray, self.jetpump_hd, self.jetpump_vd, self.jetpump_md)
        return None

    def plot_filter(self) -> None:
        """Plot the Filtered Data"""
        self._profileplot(self.hd_fit, self.vd_fit, self.md_fit, self.jetpump_hd, self.jetpump_vd, self.jetpump_md)
        return None

    def filter(self):
        """Filter WellProfile to the Minimal Data Points

        Feed the method the raw measured depth and raw true vertical depth data.
        Method will assess the raw data and represent it in the smallest number of data points.
        Method uses the segments fit function from above.

        Args:

        Returns:
            md_fit (np array): Measured Depth Filtered Data
            vd_fit (np array): Vertical Depth Filter Data
            hd_fit (np array): Horizontal Dist Filter Data
        """
        # have to use md since the valve will always be increasing
        md_fit, vd_fit = segments_fit(self.md_ray, self.vd_ray)
        md_fit[0], vd_fit[0] = 0, 0  # first values always need to start at zero
        idx = np.searchsorted(self.md_ray, md_fit)  # why is this pulling one more? because md_ray and

        hd_fit = self.hd_ray[idx]
        return hd_fit, vd_fit, md_fit

    def outflow_spacing(self, seg_len: float) -> tuple[np.ndarray, np.ndarray]:
        """Outflow Piping Spacing

        Break the outflow piping into nodes that can be fed  with piping dimension
        flowrates and etc to calculate differential pressure across them. Outflow is
        assumed to start at the jetpump.

        Args:
            seg_len (float): Segment Length of Outflow Piping, feet

        Returns:
            md_seg (np array): Measured depth Broken into segments
            vd_seg (np array): Vertical depth broken into segments
        """
        return self._outflow_spacing(self.md_fit, self.vd_fit, self.jetpump_md, seg_len)

    @staticmethod
    def _outflow_spacing(
        md_fit: np.ndarray, vd_fit: np.ndarray, outflow_md: float, seg_len: float
    ) -> tuple[np.ndarray, np.ndarray]:
        """Outflow Piping Spacing

        Break the outflow piping into nodes that can be fed  with piping dimension
        flowrates and etc to calculate differential pressure across them.

        Args:
            md_fit (np array): Filtered Measured Depth Data, feet
            vd_fit (np array): Filtered Vertical Depth Data, feet
            outflow_md (float): Depth the outflow node ends, feet
            seg_len (float): Segment Length of Outflow Piping, feet

        Returns:
            md_seg (np array): Measured depth Broken into segments
            vd_seg (np array): Vertical depth broken into segments
        """
        # need to break it up
        outflow_vd = np.interp(outflow_md, md_fit, vd_fit)
        vd_fit = vd_fit[md_fit <= outflow_md]
        md_fit = md_fit[md_fit <= outflow_md]  # keep values less than outflow_md
        md_fit = np.append(md_fit, outflow_md)  # add the final outflow md to end
        vd_fit = np.append(vd_fit, outflow_vd)
        md1 = md_fit[:-1]  # everything but the last character
        md2 = md_fit[1:]  # everything but the first character
        dist = md2 - md1  # distance between points
        md_seg = np.array([])
        for i, dis in enumerate(dist):
            # force there to always be at least three (?) spaces?
            dis = max(int(np.ceil(dis / seg_len)), 3)  # evenly space out the spaces
            md_seg = np.append(md_seg, np.linspace(md1[i], md2[i], dis))  # double counting
        md_seg = np.unique(md_seg)  # get rid of weird double counts from linspace
        vd_seg = np.interp(md_seg, md_fit, vd_fit)
        return md_seg, vd_seg

    @staticmethod
    def _depth_interp(in_dpth: float, in_ray: np.ndarray, out_ray: np.ndarray) -> float:
        """Depth Interpolation

        Args:
            in_dpth (float): Known Depth, feet
            in_ray (list): Known List of Depths, feet
            out_ray (list): Unknown List of Depths, feet

        Returns:
            out_dpth (float): Unknown Depth, Feet
        """
        if (min(in_ray) < in_dpth < max(in_ray)) is False:
            raise ValueError(f"{in_dpth} feet is not inside survey boundary")

        out_dpth = np.interp(in_dpth, in_ray, out_ray)
        return float(out_dpth)

    @staticmethod
    def _horz_dist(md_ray: np.ndarray, vd_ray: np.ndarray) -> np.ndarray:
        """Horizontal Distance from Wellhead

        Args:
            md_ray (np array): Measured Depth array, feet
            vd_ray (np array): Vertical Depth array, feet

        Returns:
            hd_ray (np array): Horizontal Dist array, feet
        """
        md_diff = np.diff(md_ray, n=1)  # difference between values in array
        vd_diff = np.diff(vd_ray, n=1)  # difference between values in array
        hd_diff = np.zeros(1)  # start with zero at top to make array match original size
        hd_diff = np.append(hd_diff, np.sqrt(md_diff**2 - vd_diff**2))  # pythagorean theorem
        hd_ray = np.cumsum(hd_diff)  # rolling sum, previous values are finite differences
        return hd_ray

    @staticmethod
    def _profileplot(
        hd_ray: np.ndarray, vd_ray: np.ndarray, md_ray: np.ndarray, hd_jp: float, vd_jp: float, md_jp: float
    ) -> None:
        """Create a Well Profile Plot

        Annotate the graph will a label of the measured depth every 1000 feet of md.
        Future note, should the hd and vd axis scales be forced to match? How hard is that?

        Args:
            hd_ray (np array): Horizontal distance, feet
            td_ray (np array): Vertical depth, feet
            md_ray (np arary): Measured depth, feet
            hd_jp (float): Horizontal distance jetpump, feet
            vd_jp (float): Vertical depth jetpump, feet
            md_jp (float): Measured depth jetpump, feet
        """
        if len(md_ray) > 20:
            plt.scatter(hd_ray, vd_ray, label="Survey")
        else:
            plt.plot(hd_ray, vd_ray, marker="o", linestyle="--", label="Survey")

        # plot jetpump location
        plt.plot(hd_jp, vd_jp, marker="o", color="r", linestyle="", label=f"Jetpump MD: {int(md_jp)} ft")

        plt.gca().invert_yaxis()
        plt.title(f"Dir Survey, Length: {max(md_ray)} ft")
        plt.xlabel("Horizontal Distance, Feet")
        plt.ylabel("True Vertical Depth, Feet")

        # find the position in the measured depth array closest to every 1000'
        md_match = np.arange(1000, max(md_ray), 1000)
        idxs = np.searchsorted(md_ray, md_match)
        idxs = np.unique(idxs)  # don't repeat values, issue on filtered data

        # annotate every ~1000' of measured depth
        for idx in idxs:
            plt.annotate(
                text=f"{int(md_ray[idx])} ft", xy=(hd_ray[idx] + 5, vd_ray[idx] - 10), rotation=30  # type: ignore
            )
        plt.legend()
        plt.axis("equal")
        plt.show()

    @classmethod
    def schrader(cls):
        """Schrader Bluff Generic Well Profile

        Generic Schrader Bluff well profile based on MPE-42 geometry.
        Imported from generic_wprof.json

        Args:
            md_list (list): MPE-42 Measured Depth
            vd_list (list): MPE-42 Vertical Depth
            jetpump_md (float): 6693 MD, feet
        """
        e42_md, e42_vd = survey_data("MPE-42")
        return cls(md_list=e42_md, vd_list=e42_vd, jetpump_md=6693)

    @classmethod
    def kuparuk(cls):
        """Kuparuk Generic Well Profile

        Generic Kuparuk well profile based on MPC-23 geometry.
        MPC-23 is a slant Kuparuk Well, so not a perfect canidate.
        Imported from generic_wprof.json

        Args:
            md_list (list): MPC-23 Measured Depth
            vd_list (list): MPC-23 Vertical Depth
            jetpump_md (float): 7926 MD, feet
        """
        c23_md, c23_vd = survey_data("MPC-23")
        return cls(md_list=c23_md, vd_list=c23_vd, jetpump_md=7926)


def sort_profile(md_ray: np.ndarray, vd_ray: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Sort Well Profile

    Take in the raw data from various databases. Sort the measured
    depth in ascending order from smallest to largest. Mirror the new
    ordered sort on the vertical array.

    Args:
        md_ray (np array): Measured Depth array, feet
        vd_ray (np array): Vertical Depth array, feet

    Returns:
        md_sort (np array): Sorted Measured Depth array, feet
        vd_sort (np array): Sorted Vertical Depth array, feet
    """
    sort_idxs = np.argsort(md_ray)
    md_sort = md_ray[sort_idxs]
    vd_sort = vd_ray[sort_idxs]
    return md_sort, vd_sort


def segments_fit(X: np.ndarray, Y: np.ndarray, maxcount: int = 18) -> tuple[np.ndarray, np.ndarray]:
    """Segments Fit DankOC

    Feed the method the raw data. Function will assess the raw data.
    Then return the smallest smallest number of data points to fit.

    Args:
        X (numpy array): Raw X Data to be fit
        Y (numpy array): Raw Y Data to be fit
        maxcount (int): Max number of points to represent the profile by

    Returns:
        X_fit (np.array): Filtered X Data
        Y_fit (np.array): Filtered Y Data

    References:
        - https://gist.github.com/ruoyu0088/70effade57483355bbd18b31dc370f2a
        - https://discovery.ucl.ac.uk/id/eprint/10070516/1/AIC_BIC_Paper.pdf
        - Comment from user: dankoc
    """
    xmin, xmax = X.min(), X.max()

    n = len(X)

    best_AIC = float("inf")
    best_BIC = float("inf")
    best_fit = None

    for count in range(1, maxcount + 1):
        seg = np.full(count - 1, (xmax - xmin) / count)

        px_init = np.r_[np.r_[xmin, seg].cumsum(), xmax]
        py_init = np.array([Y[np.abs(X - x) < (xmax - xmin) * 0.1].mean() for x in px_init])

        def func(p):
            seg = p[: count - 1]
            py = p[count - 1 :]  # noqa E203
            px = np.r_[np.r_[xmin, seg].cumsum(), xmax]
            return px, py

        def err(p):  # This is RSS / n
            px, py = func(p)
            Y2 = np.interp(X, px, py)
            return np.mean((Y - Y2) ** 2)

        res = optimize.minimize(err, x0=np.r_[seg, py_init], method="Nelder-Mead")

        # Compute AIC/ BIC.
        AIC = n * np.log(err(res.x)) + 4 * count
        BIC = n * np.log(err(res.x)) + 2 * count * np.log(n)

        if (BIC < best_BIC) & (AIC < best_AIC):  # Continue adding complexity.
            best_fit = res
            best_AIC, best_BIC = AIC, BIC
        else:  # Stop.
            count = count - 1
            break

    px, py = func(best_fit.x)  # type: ignore
    px, py = segments_guardrails(px, py, X.max(), Y.max())

    return px, py  # type: ignore [return the last (n-1)]


def segments_guardrails(px: np.ndarray, py: np.ndarray, xmax: float, ymax: float) -> tuple[np.ndarray, np.ndarray]:
    """Segments Guardrails

    Add some guardrails to ensure the returned datapoints don't go past the max values.
    Eliminate one of the values that are redundant. Maybe add something where the first
    values in the array need to be 0,0 as well? That code is repeated in other places.

    Args:
        px (np.array): Filtered x data
        py (np.array): Filtered y data
        xmax (float): Maximum x value from the non-filtered data
        ymax (float): Maximum y value from the non-filtered data

    Returns:
        px (np.array): Filtered x data with guardrails
        py (np.array): Filtered y data with guardrails
    """
    px = np.clip(px, None, xmax)  # Replace any px values longer than xmax with xmax.
    py[px == xmax] = ymax  # Any place that xmax exists in px, replace the corresponding py with ymax.

    combo = np.stack((px, py), axis=-1)  # zip together 1D into 2D that you can review
    unq_vals, unq_idxs = np.unique(combo, return_index=True, axis=0)

    px, py = px[unq_idxs], py[unq_idxs]
    return px, py


def survey_data(well_name: str) -> tuple[list, list]:
    """Survey Data

    Use the survey data from a .JSON file

    Args:
        well_name (str): Well Name

    Returns:
        md_dpth (list): List of Measured Depth, feet
        vd_dpth (list): List of Vertical Depth, feet
    """

    json_path = Path(__file__).parents[0] / "generic_wprof.json"

    # verify the well_name actually exists in the JSON file
    with open(json_path, "r") as json_file:
        survey_dict = json.load(json_file)

    # check if the name exists in the survey dictionary
    if well_name in survey_dict:
        md_dpth = survey_dict[well_name]["md_dpth"]
        vd_dpth = survey_dict[well_name]["vd_dpth"]
        return md_dpth, vd_dpth
    else:
        raise ValueError(f"Well Name {well_name} not found in the Survey Data")



import math


class Pipe:
    def __init__(self, out_dia: float, thick: float, abs_ruff: float = 0.004) -> None:
        """Initialize a Piece of Piping

        The pipe doesn't have a length. The object will work with the profile
        class to create an appropriate length and angle of piping.

        Args:
            out_dia (float): Outside diameter of the pipe, inches
            thick (float): Thickness of the piping, inches
            abs_ruff (float): Abs Roughness, inches, Default: 0.004 inches

        Returns:
            Self
        """

        if thick > out_dia:
            raise ValueError("Pipe thickness is greater than outer diameter")

        self.out_dia = out_dia
        self.thick = thick
        self.abs_ruff = abs_ruff

        self.inn_dia = out_dia - 2 * thick  # inner diamter

    def __repr__(self):
        return f"Pipe: Out Dia {self.out_dia} inches and Inn Dia {self.inn_dia} inches"

    @classmethod
    def four_half_tube(cls):
        """Four and Half inch tubing

        Generic 4.5 inch production tubing

        Args:
            out_dia (float): 4.5 inches
            thick (float): 0.271 inches
        """
        return cls(out_dia=4.5, thick=0.271)

    @classmethod
    def three_half_tube(cls):
        """Three and Half inch tubing

        Generic 3.5 inch production tubing
        """
        return cls(out_dia=3.5, thick=0.271)

    @classmethod
    def seven_case(cls):
        """Seven Inch Casing

        Generic 7.0 inch casing
        """
        return cls(out_dia=7.0, thick=0.5)

    @property
    def inn_area(self) -> float:
        """Inner area of piping, ft2"""
        return self.area_circle(self.inn_dia) / 144

    @property
    def out_area(self) -> float:
        """Outer area of piping, ft2"""
        return self.area_circle(self.out_dia) / 144

    @staticmethod
    def area_circle(diameter):
        """Area of a Circle

        Args:
            diameter (float): diameter of a circle

        Returns:
            area (float): area of a circle
        """
        area = math.pi * (diameter**2) / 4
        return area


class PipeInPipe:
    def __init__(self, inn_pipe: Pipe, out_pipe: Pipe) -> None:
        """Initialize a pipe inside another pipe

        This class allows creating a tubing and annulus combination. This class
        is used later when the well is defined as forward or reverse circulating jet pump.
        Additionally the class is used to calculate friction of the power fluid.

        Args:
            inn_pipe (Pipe): Inside Pipe in an annulus set up
            out_pipe (Pipe): Outside Pipe in an annulus set up

        Returns:
            Self
        """
        if inn_pipe.out_dia > out_pipe.inn_dia:
            raise ValueError("Inner pipe will not fit into the outer pipe")

        self.inn_pipe = inn_pipe
        self.out_pipe = out_pipe

    def __repr__(self):
        return f"Inner Pipe OD: {self.inn_pipe.out_dia} inches, Outer Pipe OD: {self.out_pipe.out_dia} inches"

    @property
    def tube_area(self) -> float:
        """Tubing Cross Sectional Area, ft2"""
        return self.inn_pipe.inn_area

    @property
    def tube_hyd_dia(self) -> float:
        """Tubing Hydraulic Diameter, inches"""
        return self.inn_pipe.inn_dia

    @property
    def tube_abs_ruff(self) -> float:
        """Tubing Absolute Roughness, inches"""
        return self.inn_pipe.abs_ruff

    @property
    def ann_area(self) -> float:
        """Annulus Cross Sectional Area, ft2"""
        return self.out_pipe.inn_area - self.inn_pipe.out_area

    @property
    def ann_hyd_dia(self) -> float:
        """Annulus Hydraulic Diameter, inches"""
        return self.out_pipe.inn_dia - self.inn_pipe.out_dia

    @property
    def ann_abs_ruff(self) -> float:
        """Annulus Absolute Roughness, inches

        Uses the same value for the inside wall of outer pipe and
        outer wall of the inside pipe
        """
        return self.out_pipe.abs_ruff



"""Forms of Mechanical Components

The following file is a place to house basic equations for calculating geometric properties.
Basic examples of the properties are the area of circle, perimeter, or angle of a right triangle.
The important configuration of these equations are they do not contain units. Units need to be
handled in other places...(is this going to get confusing?)
"""

import math

import numpy as np


def _area_circle(diameter: float) -> float:
    """Area of a Circle

    Args:
        diameter (float): diameter of a circle

    Returns:
        area (float): area of a circle
    """
    area = math.pi * (diameter**2) / 4
    return area


def vert_angle(hlen: float, ylen: float) -> float:
    """Vertical Angle, degrees

    Imagine a triangle whose hypotenuse starts at (0, 0) and ends at (x1, y1).
    The angle being calculated is between the hypotenuse and y-axis.

    Args:
        hlen (float): Length of the hypotenuse
        ylen (float): Length along the y-axis

    Returns:
        theta (float): Vertical angle, degrees
    """
    theta = np.arccos(ylen / hlen)  # angle in radians
    theta = np.degrees(theta)  # convert to degrees
    return theta


def horz_angle(hlen: float, ylen: float) -> float:
    """Horizontal Angle, degrees

    Imagine a triangle whose hypotenuse starts at (0, 0) and ends at (x1, y1).
    The angle being calculated is between the hypotenuse and x-axis.

    Args:
        hlen (float): Length of the hypotenuse
        ylen (float): Length along the y-axis

    Returns:
        theta (float): Horizontal angle, degrees
    """
    theta = np.arcsin(ylen / hlen)  # angle in radians
    theta = np.degrees(theta)  # convert to degrees
    return theta



import json
from pathlib import Path

import numpy as np
import pytest

from woffl.pvt.formgas import FormGas


def compute_formgas_data(prs_ray: np.ndarray | list, temp: float, gas_sg: float) -> dict:
    """Compute FormGas Data

    Create a list of properties of a formgas. Can be used to compare to results obtained with hysys.
    Density, viscosity and zfactor.
    """
    py_gas = FormGas(gas_sg=gas_sg)
    rho_gas, visc_gas, zfactor = [], [], []

    for prs in prs_ray:
        py_gas = py_gas.condition(prs, temp)

        rho_gas.append(py_gas.density)
        visc_gas.append(py_gas.viscosity)
        zfactor.append(py_gas.zfactor)

    pygas = {"rho_gas": rho_gas, "visc_gas": visc_gas, "zfactor": zfactor}
    return pygas


@pytest.fixture(scope="module")
def hysys_formgas():
    hysys_path = Path(__file__).parents[1] / "data" / "hysys_formgas_peng_rob.json"
    with open(hysys_path) as json_file:
        return json.load(json_file)


@pytest.fixture(scope="module")
def python_formgas(hysys_formgas):
    return compute_formgas_data(hysys_formgas["pres_psig"], hysys_formgas["temp_degf"], hysys_formgas["gas_sg"])


# blasingame 1988 PDF
blasing_ppr = 2.958
blasing_tpr = 1.867
blasing_zf = 0.9117


def test_zfactor_gradschool() -> None:
    grad_zf = FormGas._zfactor_grad_school(blasing_ppr, blasing_tpr)
    assert grad_zf == pytest.approx(blasing_zf, rel=0.01)


def test_zfactor_dranchuk() -> None:
    dak_zf = FormGas._zfactor_dak(blasing_ppr, blasing_tpr)
    assert dak_zf == pytest.approx(blasing_zf, rel=0.01)


def test_gas_compressibility() -> None:
    # properties of petroleum fluids, 2nd Edition, McCain, Pag 174
    pres_mccain = 1000 + 14.7  # psia + atm
    temp_mccain = 68
    mccain_compress = 0.001120
    methane = FormGas(gas_sg=0.55)
    calc_compress = methane.condition(pres_mccain, temp_mccain).compress
    assert calc_compress == pytest.approx(mccain_compress, rel=0.03)


def test_gas_density(hysys_formgas, python_formgas) -> None:
    np.testing.assert_allclose(hysys_formgas["rho_gas"], python_formgas["rho_gas"], rtol=0.05)


def test_gas_viscosity(hysys_formgas, python_formgas) -> None:
    np.testing.assert_allclose(hysys_formgas["visc_gas"], python_formgas["visc_gas"], rtol=0.03)


def test_gas_zfactor(hysys_formgas, python_formgas) -> None:
    np.testing.assert_allclose(hysys_formgas["zfactor"], python_formgas["zfactor"], rtol=0.04)



import pytest

from woffl.pvt import FormWater

# "test" water properties...
water = FormWater(wat_sg=1)


def test_water_density() -> None:
    assert water.density == pytest.approx(62.4)


def test_water_viscosity() -> None:
    assert water.viscosity == pytest.approx(0.75)


def test_water_compressibility() -> None:
    assert water.compress == pytest.approx(3.13e-6, rel=0.01)  # psi^-1


def test_water_tension() -> None:
    assert water.tension == pytest.approx(0.005, rel=0.01)  # lbf/ft
